"""> PyDRex: Tests for core D-Rex routines."""

import contextlib as cl
import sys

import numpy as np
import pytest
from numpy import testing as nt
from scipy.spatial.transform import Rotation

from pydrex import core as _core
from pydrex import io as _io
from pydrex import logger as _log
from pydrex import minerals as _minerals
from pydrex import utils as _utils
from pydrex import visualisation as _vis

# Subdirectory of `outdir` used to store outputs from these tests.
SUBDIR = "core"


class TestDislocationCreepOPX:
    """Single-grain orthopyroxene crystallographic rotation rate tests."""

    class_id = "dislocation_creep_OPX"

    @pytest.mark.skipif(sys.platform == "win32", reason="Not equal to tolerance")
    def test_shear_dudz(self, outdir):
        test_id = "dudz"
        optional_logging = cl.nullcontext()
        if outdir is not None:
            optional_logging = _log.logfile_enable(
                f"{outdir}/{SUBDIR}/{self.class_id}_{test_id}.log"
            )
        with optional_logging:
            for θ in np.mgrid[0 : 2 * np.pi : 360j]:
                _log.debug("θ (°): %s", np.rad2deg(θ))
                orientations_diff, fractions_diff = _core.derivatives(
                    phase=_core.MineralPhase.enstatite,
                    fabric=_core.MineralFabric.enstatite_AB,
                    n_grains=1,
                    orientations=Rotation.from_rotvec([[0, θ, 0]]).as_matrix(),
                    fractions=np.array([1.0]),
                    strain_rate=np.array([[0, 0, 1], [0, 0, 0], [1, 0, 0]]),
                    velocity_gradient=np.array([[0, 0, 2], [0, 0, 0], [0, 0, 0]]),
                    stress_exponent=1.5,
                    deformation_exponent=3.5,
                    nucleation_efficiency=5,
                    gbm_mobility=0,
                    volume_fraction=1.0,
                )
                cosθ = np.cos(θ)
                cos2θ = np.cos(2 * θ)
                sinθ = np.sin(θ)
                target_orientations_diff = (1 + cos2θ) * np.array(
                    [[sinθ, 0, -cosθ], [0, 0, 0], [cosθ, 0, sinθ]]
                )
                np.testing.assert_allclose(
                    orientations_diff[0], target_orientations_diff
                )
                assert np.isclose(np.sum(fractions_diff), 0.0)

    def test_shear_dvdx(self, outdir):
        test_id = "dvdx"
        optional_logging = cl.nullcontext()
        if outdir is not None:
            optional_logging = _log.logfile_enable(
                f"{outdir}/{SUBDIR}/{self.class_id}_{test_id}.log"
            )
        with optional_logging:
            for θ in np.linspace(0, 2 * np.pi, 361):
                _log.debug("θ (°): %s", np.rad2deg(θ))
                orientations = Rotation.from_rotvec([[0, 0, θ]]).as_matrix()
                deformation_rate = _core._get_deformation_rate(
                    _core.MineralPhase.enstatite,
                    orientations[0],
                    np.array([0, 0, 0, 0]),
                )
                np.testing.assert_allclose(deformation_rate.flatten(), np.zeros(9))
                orientations_diff, fractions_diff = _core.derivatives(
                    phase=_core.MineralPhase.enstatite,
                    fabric=_core.MineralFabric.enstatite_AB,
                    n_grains=1,
                    orientations=orientations,
                    fractions=np.array([1.0]),
                    strain_rate=np.array([[0, 1, 0], [1, 0, 0], [0, 0, 0]]),
                    velocity_gradient=np.array([[0, 0, 0], [2, 0, 0], [0, 0, 0]]),
                    stress_exponent=1.5,
                    deformation_exponent=3.5,
                    nucleation_efficiency=5,
                    gbm_mobility=0,
                    volume_fraction=1.0,
                )
                # Can't activate the (100)[001] slip system, no plastic deformation.
                # Only passive (rigid-body) rotation due to the velocity gradient.
                sinθ = np.sin(θ)
                cosθ = np.cos(θ)
                target_orientations_diff = np.array(
                    [[sinθ, cosθ, 0], [-cosθ, sinθ, 0], [0, 0, 0]]
                )
                np.testing.assert_allclose(
                    orientations_diff[0], target_orientations_diff, atol=1e-15
                )
                assert np.isclose(np.sum(fractions_diff), 0.0)


class TestDislocationCreepOlivineA:
    """Single-grain A-type olivine analytical rotation rate tests."""

    class_id = "dislocation_creep_OlA"

    @pytest.mark.skipif(sys.platform == "win32", reason="Not equal to tolerance")
    def test_shear_dvdx_slip_010_100(self, outdir):
        r"""Single grain of A-type olivine, slip on (010)[100].

        Velocity gradient:
        $$\bm{L} = \begin{bmatrix} 0 & 0 & 0 \cr 2 & 0 & 0 \cr 0 & 0 & 0 \end{bmatrix}$$

        """
        test_id = "dvdx_010_100"
        nondim_velocity_gradient = np.array([[0, 0, 0], [2, 0, 0], [0, 0, 0]])
        # Strain rate is 0.5*(L + Lᵀ).
        nondim_strain_rate = np.array([[0, 1, 0], [1, 0, 0], [0, 0, 0]])

        optional_logging = cl.nullcontext()
        if outdir is not None:
            optional_logging = _log.logfile_enable(
                f"{outdir}/{SUBDIR}/{self.class_id}_{test_id}.log"
            )
            initial_angles = []
            rotation_rates = []
            target_rotation_rates = []

        with optional_logging:
            for θ in np.mgrid[0 : 2 * np.pi : 3600j]:
                _log.debug("θ (°): %s", np.rad2deg(θ))
                # Initial grain rotations around Z (anti-clockwise).
                initial_orientations = Rotation.from_rotvec([[0, 0, θ]])
                orientation_matrix = initial_orientations[0].as_matrix()
                slip_invariants = _core._get_slip_invariants(
                    nondim_strain_rate, orientation_matrix
                )
                _log.debug("slip invariants: %s", slip_invariants)

                crss = _core.get_crss(
                    _core.MineralPhase.olivine,
                    _core.MineralFabric.olivine_A,
                )
                slip_indices = np.argsort(np.abs(slip_invariants / crss))
                slip_system = _minerals.OLIVINE_SLIP_SYSTEMS[slip_indices[-1]]
                assert slip_system == ([0, 1, 0], [1, 0, 0])

                slip_rates = _core._get_slip_rates_olivine(
                    slip_invariants, slip_indices, crss, 3.5
                )
                _log.debug("slip rates: %s", slip_rates)

                deformation_rate = _core._get_deformation_rate(
                    _core.MineralPhase.olivine,
                    orientation_matrix,
                    slip_rates,
                )
                _log.debug("deformation rate:\n%s", deformation_rate)

                orientations_diff, fractions_diff = _core.derivatives(
                    phase=_core.MineralPhase.olivine,
                    fabric=_core.MineralFabric.olivine_A,
                    n_grains=1,
                    orientations=initial_orientations.as_matrix(),
                    fractions=np.array([1.0]),
                    strain_rate=nondim_strain_rate,
                    velocity_gradient=nondim_velocity_gradient,
                    stress_exponent=1.5,
                    deformation_exponent=3.5,
                    nucleation_efficiency=5,
                    gbm_mobility=0,
                    volume_fraction=1.0,
                )
                cosθ = np.cos(θ)
                cos2θ = np.cos(2 * θ)
                sinθ = np.sin(θ)
                target_orientations_diff = np.array(
                    [
                        [sinθ * (1 + cos2θ), cosθ * (1 + cos2θ), 0],
                        [-cosθ * (1 + cos2θ), sinθ * (1 + cos2θ), 0],
                        [0, 0, 0],
                    ]
                )
                np.testing.assert_allclose(
                    orientations_diff[0], target_orientations_diff
                )
                if outdir is not None:
                    initial_angles.append(np.rad2deg(θ))
                    rotation_rates.append(
                        np.sqrt(
                            orientations_diff[0][0, 0] ** 2
                            + orientations_diff[0][0, 1] ** 2
                        )
                    )
                    target_rotation_rates.append(1 + cos2θ)
                assert np.isclose(np.sum(fractions_diff), 0.0)

        if outdir is not None:
            fig = _vis.figure(figscale=(1, 2 / 3))
            axs = fig.subplot_mosaic(  # Put the legend on a special axes.
                [["legend"], ["spin"]], gridspec_kw={"height_ratios": [0.1, 1]}
            )
            ax = axs["spin"]
            fig, ax, colors = _vis.spin(
                ax,
                initial_angles[::25],
                rotation_rates[::25],
                initial_angles,
                target_rotation_rates,
                shear_axis=90,
            )
            _utils.redraw_legend(
                ax,
                legendax=axs["legend"],
                loc="upper center",
                ncols=3,
                mode="expand",
            )
            fig.savefig(
                _io.resolve_path(f"{outdir}/{SUBDIR}/{self.class_id}_{test_id}.pdf")
            )

    def test_shear_dudz_slip_001_100(self, outdir):
        r"""Single grain of A-type olivine, slip on (001)[100].

        Velocity gradient:
        $$\bm{L} = \begin{bmatrix} 0 & 0 & 2 \cr 0 & 0 & 0 \cr 0 & 0 & 0 \end{bmatrix}$$

        """
        test_id = "dudz_001_100"
        nondim_velocity_gradient = np.array([[0, 0, 2], [0, 0, 0], [0, 0, 0]])
        # Strain rate is 0.5*(L + Lᵀ).
        nondim_strain_rate = np.array([[0, 0, 1], [0, 0, 0], [1, 0, 0]])

        optional_logging = cl.nullcontext()
        if outdir is not None:
            optional_logging = _log.logfile_enable(
                f"{outdir}/{SUBDIR}/{self.class_id}_{test_id}.log"
            )
            initial_angles = []
            rotation_rates = []
            target_rotation_rates = []

        with optional_logging:
            for θ in np.mgrid[0 : 2 * np.pi : 360j]:
                _log.debug("θ (°): %s", np.rad2deg(θ))
                # Initial grain rotations around Y (anti-clockwise).
                initial_orientations = Rotation.from_rotvec([[0, θ, 0]])
                orientation_matrix = initial_orientations[0].as_matrix()
                slip_invariants = _core._get_slip_invariants(
                    nondim_strain_rate, orientation_matrix
                )
                _log.debug("slip invariants: %s", slip_invariants)

                crss = _core.get_crss(
                    _core.MineralPhase.olivine,
                    _core.MineralFabric.olivine_A,
                )
                slip_indices = np.argsort(np.abs(slip_invariants / crss))
                slip_system = _minerals.OLIVINE_SLIP_SYSTEMS[slip_indices[-1]]
                assert slip_system == ([0, 0, 1], [1, 0, 0])

                slip_rates = _core._get_slip_rates_olivine(
                    slip_invariants, slip_indices, crss, 3.5
                )
                _log.debug("slip rates: %s", slip_rates)

                deformation_rate = _core._get_deformation_rate(
                    _core.MineralPhase.olivine,
                    orientation_matrix,
                    slip_rates,
                )
                _log.debug("deformation rate:\n%s", deformation_rate)

                orientations_diff, fractions_diff = _core.derivatives(
                    phase=_core.MineralPhase.olivine,
                    fabric=_core.MineralFabric.olivine_A,
                    n_grains=1,
                    orientations=initial_orientations.as_matrix(),
                    fractions=np.array([1.0]),
                    strain_rate=nondim_strain_rate,
                    velocity_gradient=nondim_velocity_gradient,
                    stress_exponent=1.5,
                    deformation_exponent=3.5,
                    nucleation_efficiency=5,
                    gbm_mobility=0,
                    volume_fraction=1.0,
                )
                cosθ = np.cos(θ)
                cos2θ = np.cos(2 * θ)
                sinθ = np.sin(θ)
                target_orientations_diff = np.array(
                    [
                        [-sinθ * (cos2θ - 1), 0, cosθ * (cos2θ - 1)],
                        [0, 0, 0],
                        [cosθ * (1 - cos2θ), 0, sinθ * (1 - cos2θ)],
                    ]
                )
                np.testing.assert_allclose(
                    orientations_diff[0], target_orientations_diff
                )
                if outdir is not None:
                    initial_angles.append(np.rad2deg(θ))
                    rotation_rates.append(
                        np.sqrt(
                            orientations_diff[0][0, 0] ** 2
                            + orientations_diff[0][0, 2] ** 2
                        )
                    )
                    target_rotation_rates.append(1 - cos2θ)
                assert np.isclose(np.sum(fractions_diff), 0.0)

        if outdir is not None:
            fig = _vis.figure(figscale=(1, 2 / 3))
            axs = fig.subplot_mosaic(  # Put the legend on a special axes.
                [["legend"], ["spin"]], gridspec_kw={"height_ratios": [0.1, 1]}
            )
            ax = axs["spin"]
            fig, ax, colors = _vis.spin(
                ax,
                initial_angles[::5],
                rotation_rates[::5],
                initial_angles,
                target_rotation_rates,
                shear_axis=0,
            )
            _utils.redraw_legend(
                ax,
                legendax=axs["legend"],
                loc="upper center",
                ncols=3,
                mode="expand",
            )
            fig.savefig(
                _io.resolve_path(f"{outdir}/{SUBDIR}/{self.class_id}_{test_id}.pdf")
            )

    @pytest.mark.skipif(sys.platform == "win32", reason="Not equal to tolerance")
    def test_shear_dwdx_slip_001_100(self, outdir):
        r"""Single grain of A-type olivine, slip on (001)[100].

        Velocity gradient:
        $$\bm{L} = \begin{bmatrix} 0 & 0 & 0 \cr 0 & 0 & 0 \cr 2 & 0 & 0 \end{bmatrix}$$

        """
        test_id = "dwdx_001_100"
        nondim_velocity_gradient = np.array([[0, 0, 0], [0, 0, 0], [2, 0, 0]])
        # Strain rate is 0.5*(L + Lᵀ).
        nondim_strain_rate = np.array([[0, 0, 1], [0, 0, 0], [1, 0, 0]])

        optional_logging = cl.nullcontext()
        if outdir is not None:
            optional_logging = _log.logfile_enable(
                f"{outdir}/{SUBDIR}/{self.class_id}_{test_id}.log"
            )
            initial_angles = []
            rotation_rates = []
            target_rotation_rates = []

        with optional_logging:
            for θ in np.mgrid[0 : 2 * np.pi : 360j]:
                _log.debug("θ (°): %s", np.rad2deg(θ))
                # Initial grain rotations around Y (anti-clockwise).
                initial_orientations = Rotation.from_rotvec([[0, θ, 0]])
                orientation_matrix = initial_orientations[0].as_matrix()
                slip_invariants = _core._get_slip_invariants(
                    nondim_strain_rate, orientation_matrix
                )
                _log.debug("slip invariants: %s", slip_invariants)

                crss = _core.get_crss(
                    _core.MineralPhase.olivine,
                    _core.MineralFabric.olivine_A,
                )
                slip_indices = np.argsort(np.abs(slip_invariants / crss))
                slip_system = _minerals.OLIVINE_SLIP_SYSTEMS[slip_indices[-1]]
                assert slip_system == ([0, 0, 1], [1, 0, 0])

                slip_rates = _core._get_slip_rates_olivine(
                    slip_invariants, slip_indices, crss, 3.5
                )
                _log.debug("slip rates: %s", slip_rates)

                deformation_rate = _core._get_deformation_rate(
                    _core.MineralPhase.olivine,
                    orientation_matrix,
                    slip_rates,
                )
                _log.debug("deformation rate:\n%s", deformation_rate)

                orientations_diff, fractions_diff = _core.derivatives(
                    phase=_core.MineralPhase.olivine,
                    fabric=_core.MineralFabric.olivine_A,
                    n_grains=1,
                    orientations=initial_orientations.as_matrix(),
                    fractions=np.array([1.0]),
                    strain_rate=nondim_strain_rate,
                    velocity_gradient=nondim_velocity_gradient,
                    stress_exponent=1.5,
                    deformation_exponent=3.5,
                    nucleation_efficiency=5,
                    gbm_mobility=0,
                    volume_fraction=1.0,
                )
                cosθ = np.cos(θ)
                cos2θ = np.cos(2 * θ)
                sinθ = np.sin(θ)
                target_orientations_diff = np.array(
                    [
                        [-sinθ * (1 + cos2θ), 0, cosθ * (1 + cos2θ)],
                        [0, 0, 0],
                        [-cosθ * (1 + cos2θ), 0, -sinθ * (1 + cos2θ)],
                    ]
                )
                np.testing.assert_allclose(
                    orientations_diff[0], target_orientations_diff
                )
                if outdir is not None:
                    initial_angles.append(np.rad2deg(θ))
                    rotation_rates.append(
                        np.sqrt(
                            orientations_diff[0][0, 0] ** 2
                            + orientations_diff[0][0, 2] ** 2
                        )
                    )
                    target_rotation_rates.append(1 + cos2θ)
                assert np.isclose(np.sum(fractions_diff), 0.0)

        if outdir is not None:
            fig = _vis.figure(figscale=(1, 2 / 3))
            axs = fig.subplot_mosaic(  # Put the legend on a special axes.
                [["legend"], ["spin"]], gridspec_kw={"height_ratios": [0.1, 1]}
            )
            ax = axs["spin"]
            fig, ax, colors = _vis.spin(
                ax,
                initial_angles[::5],
                rotation_rates[::5],
                initial_angles,
                target_rotation_rates,
                shear_axis=90,
            )
            _utils.redraw_legend(
                ax,
                legendax=axs["legend"],
                loc="upper center",
                ncols=3,
                mode="expand",
            )
            fig.savefig(
                _io.resolve_path(f"{outdir}/{SUBDIR}/{self.class_id}_{test_id}.pdf")
            )

    @pytest.mark.skipif(sys.platform == "win32", reason="Not equal to tolerance")
    def test_shear_dvdz_slip_010_001(self, outdir):
        r"""Single grain of A-type olivine, slip on (010)[001].

        Velocity gradient:
        $$\bm{L} = \begin{bmatrix} 0 & 0 & 0 \cr 0 & 0 & 2 \cr 0 & 0 & 0 \end{bmatrix}$$

        """
        test_id = "dvdz_010_001"
        nondim_velocity_gradient = np.array([[0, 0, 0], [0, 0, 2], [0, 0, 0]])
        # Strain rate is 0.5*(L + Lᵀ).
        nondim_strain_rate = np.array([[0, 0, 0], [0, 0, 1], [0, 1, 0]])

        optional_logging = cl.nullcontext()
        if outdir is not None:
            optional_logging = _log.logfile_enable(
                f"{outdir}/{SUBDIR}/{self.class_id}_{test_id}.log"
            )
            initial_angles = []
            rotation_rates = []
            target_rotation_rates = []

        with optional_logging:
            for θ in np.mgrid[0 : 2 * np.pi : 360j]:
                _log.debug("θ (°): %s", np.rad2deg(θ))
                # Initial grain rotations around X (anti-clockwise).
                initial_orientations = Rotation.from_rotvec([[θ, 0, 0]])
                orientation_matrix = initial_orientations[0].as_matrix()
                slip_invariants = _core._get_slip_invariants(
                    nondim_strain_rate, orientation_matrix
                )
                _log.debug("slip invariants: %s", slip_invariants)

                crss = _core.get_crss(
                    _core.MineralPhase.olivine,
                    _core.MineralFabric.olivine_A,
                )
                slip_indices = np.argsort(np.abs(slip_invariants / crss))
                slip_system = _minerals.OLIVINE_SLIP_SYSTEMS[slip_indices[-1]]
                assert slip_system == ([0, 1, 0], [0, 0, 1])

                slip_rates = _core._get_slip_rates_olivine(
                    slip_invariants, slip_indices, crss, 3.5
                )
                _log.debug("slip rates: %s", slip_rates)

                deformation_rate = _core._get_deformation_rate(
                    _core.MineralPhase.olivine,
                    orientation_matrix,
                    slip_rates,
                )
                _log.debug("deformation rate:\n%s", deformation_rate)

                orientations_diff, fractions_diff = _core.derivatives(
                    phase=_core.MineralPhase.olivine,
                    fabric=_core.MineralFabric.olivine_A,
                    n_grains=1,
                    orientations=initial_orientations.as_matrix(),
                    fractions=np.array([1.0]),
                    strain_rate=nondim_strain_rate,
                    velocity_gradient=nondim_velocity_gradient,
                    stress_exponent=1.5,
                    deformation_exponent=3.5,
                    nucleation_efficiency=5,
                    gbm_mobility=0,
                    volume_fraction=1.0,
                )
                cosθ = np.cos(θ)
                cos2θ = np.cos(2 * θ)
                sinθ = np.sin(θ)
                target_orientations_diff = np.array(
                    [
                        [0, 0, 0],
                        [0, -sinθ * (1 + cos2θ), -cosθ * (1 + cos2θ)],
                        [0, cosθ * (1 + cos2θ), -sinθ * (1 + cos2θ)],
                    ]
                )
                np.testing.assert_allclose(
                    orientations_diff[0], target_orientations_diff
                )
                if outdir is not None:
                    initial_angles.append(np.rad2deg(θ))
                    rotation_rates.append(
                        np.sqrt(
                            orientations_diff[0][1, 1] ** 2
                            + orientations_diff[0][1, 2] ** 2
                        )
                    )
                    target_rotation_rates.append(1 + cos2θ)
                assert np.isclose(np.sum(fractions_diff), 0.0)

        if outdir is not None:
            fig = _vis.figure(figscale=(1, 2 / 3))
            axs = fig.subplot_mosaic(  # Put the legend on a special axes.
                [["legend"], ["spin"]], gridspec_kw={"height_ratios": [0.1, 1]}
            )
            ax = axs["spin"]
            fig, ax, colors = _vis.spin(
                ax,
                initial_angles[::5],
                rotation_rates[::5],
                initial_angles,
                target_rotation_rates,
                shear_axis=90,
            )
            _utils.redraw_legend(
                ax,
                legendax=axs["legend"],
                loc="upper center",
                ncols=3,
                mode="expand",
            )
            fig.savefig(
                _io.resolve_path(f"{outdir}/{SUBDIR}/{self.class_id}_{test_id}.pdf")
            )


class TestRecrystallisation2D:
    """Basic recrystallisation tests for 2D simple shear."""

    class_id = "recrystallisation_2D"

    def test_shear_dvdx_circle_inplane(self, outdir):
        r"""360000 grains of A-type olivine with uniform spread of a-axes on a circle.

        Grain growth rates are compared to analytical calculations.
        The a-axes are distributed in the YX plane (i.e.\ rotated around Z).

        Velocity gradient:
        $$\bm{L} = \begin{bmatrix} 0 & 0 & 0 \cr 2 & 0 & 0 \cr 0 & 0 & 0 \end{bmatrix}$$

        """
        test_id = "dvdx_circle_inplane"

        optional_logging = cl.nullcontext()
        # Initial uniform distribution of orientations on a 2D circle.
        initial_angles = np.mgrid[0 : 2 * np.pi : 360000j]
        cos2θ = np.cos(2 * initial_angles)
        if outdir is not None:
            out_basepath = f"{outdir}/{SUBDIR}/{self.class_id}_{test_id}"
            optional_logging = _log.logfile_enable(f"{out_basepath}.log")

        with optional_logging:
            initial_orientations = Rotation.from_rotvec(
                [[0, 0, θ] for θ in initial_angles]
            )
            orientations_diff, fractions_diff = _core.derivatives(
                phase=_core.MineralPhase.olivine,
                fabric=_core.MineralFabric.olivine_A,
                n_grains=360000,
                orientations=initial_orientations.as_matrix(),
                fractions=np.full(360000, 1 / 360000),
                strain_rate=np.array([[0, 1, 0], [1, 0, 0], [0, 0, 0]]),
                velocity_gradient=np.array([[0, 0, 0], [2, 0, 0], [0, 0, 0]]),
                stress_exponent=1.5,
                deformation_exponent=3.5,
                nucleation_efficiency=5,
                gbm_mobility=125,
                volume_fraction=1.0,
            )
            ρ = np.abs(cos2θ) ** (1.5 / 3.5)
            # No need to sum over slip systems, only (010)[100] can activate in this
            # 2D simple shear deformation geometry (ρₛ=ρ).
            target_strain_energies = ρ * np.exp(-5 * ρ**2)
            target_fractions_diff = np.array(
                [  # df/dt* = - M* f (E - E_mean)
                    -125 * 1 / 360000 * (E - np.mean(target_strain_energies))
                    for E in target_strain_energies
                ]
            )

        if outdir is not None:
            fig = _vis.figure(figscale=(1, 4 / 3))
            axs = fig.subplot_mosaic(  # Put the legend on a special axes.
                [["legend"], ["spin"], ["growth"]],
                gridspec_kw={"height_ratios": [0.2, 1, 1]},
                sharex=True,
            )
            ax = axs["spin"]
            xvals = np.rad2deg(initial_angles)
            fig, ax, colors = _vis.spin(
                ax,
                xvals[::2500],
                np.sqrt(
                    [
                        o[0, 0] ** 2 + o[0, 1] ** 2 + o[0, 2] ** 2
                        for o in orientations_diff
                    ][::2500]
                ),
                xvals,
                1 + cos2θ,
                shear_axis=90,
            )
            ax.get_legend().remove()
            ax.label_outer()
            ax2 = axs["growth"]
            fig, ax2, colors = _vis.growth(
                ax2,
                xvals[::2500],
                fractions_diff[::2500],
                xvals,
                target_fractions_diff,
                shear_axis=90,
            )
            _utils.redraw_legend(
                ax2,
                legendax=axs["legend"],
                loc="upper center",
                ncols=3,
                mode="expand",
            )
            _utils.add_subplot_labels(axs, labelmap={"spin": "a)", "growth": "b)"})
            fig.savefig(f"{out_basepath}.pdf")

        nt.assert_allclose(fractions_diff, target_fractions_diff, atol=1e-15, rtol=0)

    def test_shear_dvdx_circle_shearplane(self, outdir):
        r"""360000 grains of A-type olivine with uniform spread of a-axes on a circle.

        Unlike `test_shear_dvdx_circle_inplane`, two slip systems are active here,
        with cyclical variety in which one is dominant depending on grain orientation.
        The a-axes are distributed in the YZ plane
        (i.e.\ extrinsic rotation around Z by 90° and then around X).

        Velocity gradient:
        $$\bm{L} = \begin{bmatrix} 0 & 0 & 0 \cr 2 & 0 & 0 \cr 0 & 0 & 0 \end{bmatrix}$$

        """
        test_id = "dvdx_circle_shearplane"

        optional_logging = cl.nullcontext()
        # Initial uniform distribution of orientations on a 2D circle.
        initial_angles = np.mgrid[0 : 2 * np.pi : 360000j]
        if outdir is not None:
            out_basepath = f"{outdir}/{SUBDIR}/{self.class_id}_{test_id}"
            optional_logging = _log.logfile_enable(f"{out_basepath}.log")

        with optional_logging:
            initial_orientations = Rotation.from_euler(
                "zx", [[np.pi / 2, θ] for θ in initial_angles]
            )
            orientations_diff, fractions_diff = _core.derivatives(
                phase=_core.MineralPhase.olivine,
                fabric=_core.MineralFabric.olivine_A,
                n_grains=360000,
                orientations=initial_orientations.as_matrix(),
                fractions=np.full(360000, 1 / 360000),
                strain_rate=np.array([[0, 1, 0], [1, 0, 0], [0, 0, 0]]),
                velocity_gradient=np.array([[0, 0, 0], [2, 0, 0], [0, 0, 0]]),
                stress_exponent=1.5,
                deformation_exponent=3.5,
                nucleation_efficiency=5,
                gbm_mobility=125,
                volume_fraction=1.0,
            )

        if outdir is not None:
            fig = _vis.figure(figscale=(1, 4 / 3))
            axs = fig.subplot_mosaic([["spin"], ["growth"]], sharex=True)
            ax = axs["spin"]
            xvals = np.rad2deg(initial_angles)
            fig, ax, colors = _vis.spin(
                ax,
                xvals[::2500],
                np.sqrt(
                    [
                        o[0, 0] ** 2 + o[0, 1] ** 2 + o[0, 2] ** 2
                        for o in orientations_diff
                    ][::2500]
                ),
            )
            ax.get_legend().remove()
            ax.label_outer()
            ax2 = axs["growth"]
            fig, ax2, colors = _vis.growth(ax2, xvals[::2500], fractions_diff[::2500])
            _utils.add_subplot_labels(axs, labelmap={"spin": "a)", "growth": "b)"})
            fig.savefig(f"{out_basepath}.pdf")

        # Check dominant slip system every 1°.
        for θ in initial_angles[::1000]:
            slip_invariants = _core._get_slip_invariants(
                np.array([[0, 1, 0], [1, 0, 0], [0, 0, 0]]),
                Rotation.from_euler("zx", [np.pi / 2, θ]).as_matrix(),
            )
            θ = np.rad2deg(θ)
            crss = _core.get_crss(
                _core.MineralPhase.olivine,
                _core.MineralFabric.olivine_A,
            )
            slip_indices = np.argsort(np.abs(slip_invariants / crss))
            slip_system = _minerals.OLIVINE_SLIP_SYSTEMS[slip_indices[-1]]

            if 0 <= θ < 64:
                assert slip_system == ([0, 1, 0], [1, 0, 0])
            elif 64 <= θ < 117:
                assert slip_system == ([0, 0, 1], [1, 0, 0])
            elif 117 <= θ < 244:
                assert slip_system == ([0, 1, 0], [1, 0, 0])
            elif 244 <= θ < 297:
                assert slip_system == ([0, 0, 1], [1, 0, 0])
            else:
                assert slip_system == ([0, 1, 0], [1, 0, 0])



"""> PyDRex: Simple shear 3D tests."""

import contextlib as cl
import functools as ft
from time import process_time

import numpy as np
import pytest

from pydrex import core as _core
from pydrex import diagnostics as _diagnostics
from pydrex import geometry as _geo
from pydrex import io as _io
from pydrex import logger as _log
from pydrex import minerals as _minerals
from pydrex import stats as _stats
from pydrex import utils as _utils
from pydrex import velocity as _velocity

Pool, HAS_RAY = _utils.import_proc_pool()
if HAS_RAY:
    import ray

    from pydrex import distributed as _dstr

# Subdirectory of `outdir` used to store outputs from these tests.
SUBDIR = "3d_simple_shear"


class TestFraters2021:
    """Tests inspired by the benchmarks presented in [Fraters & Billen, 2021].

    [Fraters & Billen, 2021]: https://doi.org/10.1029/2021GC009846

    """

    class_id = "Fraters2021"

    @classmethod
    def run(
        cls,
        params,
        timestamps,
        get_velocity_gradient_initial,
        get_velocity_gradient_final,
        switch_time,
        msg,
        seed=None,
    ):
        """Run simulation with stationary particles in the given velocity gradient.

        The optional RNG `seed` is used for the initial pseudorandom orientations.
        A prefix `msg` will be printed before each timestep log message if given.
        Other keyword args are passed to `pydrex.Mineral.update_orientations`.

        Returns a tuple containing one olivine (A-type) and one enstatite mineral.
        If `params["enstatite_fraction"]` is zero, then the second tuple element will be
        `None` instead.

        """

        def get_position(t):
            return np.full(3, np.nan)

        olivine = _minerals.Mineral(
            phase=_core.MineralPhase.olivine,
            fabric=_core.MineralFabric.olivine_A,
            regime=_core.DeformationRegime.dislocation,
            n_grains=params["number_of_grains"],
            seed=seed,
        )
        if params["enstatite_fraction"] > 0:
            enstatite = _minerals.Mineral(
                phase=_core.MineralPhase.enstatite,
                fabric=_core.MineralFabric.enstatite_AB,
                regime=_core.DeformationRegime.dislocation,
                n_grains=params["number_of_grains"],
                seed=seed,
            )
        else:
            enstatite = None
        deformation_gradient = np.eye(3)  # Undeformed initial state.

        for t, time in enumerate(timestamps[:-1], start=1):
            _log.info(
                "%s; # %d; step %d/%d (t = %s)", msg, seed, t, len(timestamps) - 1, time
            )
            if time > switch_time:
                get_velocity_gradient = get_velocity_gradient_final
            else:
                get_velocity_gradient = get_velocity_gradient_initial

            if params["enstatite_fraction"] > 0:
                enstatite.update_orientations(
                    params,
                    deformation_gradient,
                    get_velocity_gradient,
                    pathline=(time, timestamps[t], get_position),
                )
            deformation_gradient = olivine.update_orientations(
                params,
                deformation_gradient,
                get_velocity_gradient,
                pathline=(time, timestamps[t], get_position),
            )
            _log.debug(
                "› velocity gradient = %s",
                get_velocity_gradient(np.nan, np.full(3, np.nan)).flatten(),
            )
        return olivine, enstatite

    @pytest.mark.slow
    @pytest.mark.parametrize("switch_time_Ma", [0, 1, 2.5, np.inf])
    def test_direction_change(
        self, outdir, seeds, params_Fraters2021, switch_time_Ma, ncpus, ray_session
    ):
        """Test a-axis alignment in simple shear with instantaneous geometry change.

        The simulation runs for 5 Ma with a strain rate of 1.58e-14/s, resulting in an
        accumulated strain invariant of 2.5.

        The initial shear has nonzero du/dz and the final shear has nonzero dv/dx where
        u is the velocity along X and v the velocity along Y.

        """
        # Strain rate in units of strain per year, avoids tiny numbers ∼1e-14.
        strain_rate = 5e-7
        # With 500 steps, each step is ~3e11 seconds which is about as small as
        # geodynamic modelling could feasibly go in most cases.
        n_timestamps = 500
        # Solve until D₀t=2.5 ('shear' γ=5).
        timestamps = np.linspace(0, 5e6, n_timestamps)
        _seeds = seeds[:500]  # 500 runs as per Fraters & Billen, 2021, Figure 5.
        n_seeds = len(_seeds)

        _, get_velocity_gradient_initial = _velocity.simple_shear_2d(
            "X", "Z", strain_rate
        )
        _, get_velocity_gradient_final = _velocity.simple_shear_2d(
            "Y", "X", strain_rate
        )

        # Output arrays for mean [100] angles.
        olA_from_proj_XZ = np.empty((n_seeds, n_timestamps))
        olA_from_proj_YX = np.empty_like(olA_from_proj_XZ)
        ens_from_proj_XZ = np.empty_like(olA_from_proj_XZ)
        ens_from_proj_YX = np.empty_like(olA_from_proj_XZ)
        # Output arrays for M-index (CPO strength).
        olA_strength = np.empty_like(olA_from_proj_XZ)
        ens_strength = np.empty_like(olA_from_proj_XZ)

        _id = _io.stringify(switch_time_Ma * 1e6)
        optional_logging = cl.nullcontext()
        if outdir is not None:
            out_basepath = f"{outdir}/{SUBDIR}/{self.class_id}_direction_change_{_id}"
            optional_logging = _log.logfile_enable(f"{out_basepath}.log")

        with optional_logging:
            clock_start = process_time()
            _run = ft.partial(
                self.run,
                params_Fraters2021,
                timestamps,
                get_velocity_gradient_initial,
                get_velocity_gradient_final,
                switch_time_Ma * 1e6,
                _id,
            )
            with Pool(processes=ncpus) as pool:
                for s, out in enumerate(pool.imap_unordered(_run, _seeds)):
                    olivine, enstatite = out
                    _log.info("%s; # %d; postprocessing olivine...", _id, _seeds[s])
                    olA_resampled, _ = _stats.resample_orientations(
                        olivine.orientations, olivine.fractions, seed=_seeds[s]
                    )
                    olA_mean_vectors = np.array(
                        [
                            _diagnostics.bingham_average(dcm, axis="a")
                            for dcm in olA_resampled
                        ]
                    )
                    olA_from_proj_XZ[s, :] = np.array(
                        [
                            _diagnostics.smallest_angle(v, v - v * [0, 1, 0])
                            for v in olA_mean_vectors
                        ]
                    )
                    olA_from_proj_YX[s, :] = np.array(
                        [
                            _diagnostics.smallest_angle(v, v - v * [0, 0, 1])
                            for v in olA_mean_vectors
                        ]
                    )
                    olA_downsampled, _ = _stats.resample_orientations(
                        olivine.orientations, olivine.fractions, seed=_seeds[s], n_samples=1000
                    )
                    olA_strength[s, :] = _diagnostics.misorientation_indices(
                        olA_downsampled, _geo.LatticeSystem.orthorhombic, pool=pool
                    )

                    del olivine, olA_resampled, olA_mean_vectors

                    _log.info("%s; # %d; postprocessing enstatite...", _id, _seeds[s])
                    ens_resampled, _ = _stats.resample_orientations(
                        enstatite.orientations, enstatite.fractions, seed=_seeds[s]
                    )
                    ens_mean_vectors = np.array(
                        [
                            _diagnostics.bingham_average(dcm, axis="a")
                            for dcm in ens_resampled
                        ]
                    )
                    ens_from_proj_XZ[s, :] = np.array(
                        [
                            _diagnostics.smallest_angle(v, v - v * [0, 1, 0])
                            for v in ens_mean_vectors
                        ]
                    )
                    ens_from_proj_YX[s, :] = np.array(
                        [
                            _diagnostics.smallest_angle(v, v - v * [0, 0, 1])
                            for v in ens_mean_vectors
                        ]
                    )
                    ens_downsampled, _ = _stats.resample_orientations(
                        enstatite.orientations, enstatite.fractions, seed=_seeds[s], n_samples=1000
                    )
                    ens_strength[s, :] = _diagnostics.misorientation_indices(
                        ens_downsampled, _geo.LatticeSystem.orthorhombic, pool=pool
                    )
                    del enstatite, ens_resampled, ens_mean_vectors

            _log.info("elapsed CPU time: %s", np.abs(process_time() - clock_start))
            _log.info("calculating ensemble averages...")
            olA_from_proj_XZ_mean = olA_from_proj_XZ.mean(axis=0)
            olA_from_proj_XZ_err = olA_from_proj_XZ.std(axis=0)

            olA_from_proj_YX_mean = olA_from_proj_YX.mean(axis=0)
            olA_from_proj_YX_err = olA_from_proj_YX.std(axis=0)

            ens_from_proj_XZ_mean = ens_from_proj_XZ.mean(axis=0)
            ens_from_proj_XZ_err = ens_from_proj_XZ.std(axis=0)

            ens_from_proj_YX_mean = ens_from_proj_YX.mean(axis=0)
            ens_from_proj_YX_err = ens_from_proj_YX.std(axis=0)

            if outdir is not None:
                np.savez(
                    f"{out_basepath}.npz",
                    olA_from_proj_XZ_mean=olA_from_proj_XZ_mean,
                    olA_from_proj_XZ_err=olA_from_proj_XZ_err,
                    olA_from_proj_YX_mean=olA_from_proj_YX_mean,
                    olA_from_proj_YX_err=olA_from_proj_YX_err,
                    ens_from_proj_XZ_mean=ens_from_proj_XZ_mean,
                    ens_from_proj_XZ_err=ens_from_proj_XZ_err,
                    ens_from_proj_YX_mean=ens_from_proj_YX_mean,
                    ens_from_proj_YX_err=ens_from_proj_YX_err,
                )
                np.savez(
                    f"{out_basepath}_strength.npz",
                    olA_mean=olA_strength.mean(axis=0),
                    ens_mean=ens_strength.mean(axis=0),
                    olA_err=olA_strength.std(axis=0),
                    ens_err=ens_strength.std(axis=0),
                )



"""> PyDRex: Builtin mesh generation helpers using meshio and pygmsh."""

from dataclasses import dataclass, field

import gmsh as gm
import numpy as np

from pydrex import exceptions as _err
from pydrex import geometry as _geo
from pydrex import logger as _log


@dataclass
class Model:
    """A context manager for using the gmsh model API.

    Examples:

    >>> with Model("example_model", 2, _write_file=False) as model:
    ...     model.point_constraints = [
    ...         (0, -10, 0, 0.1),  # x, y, z, nearby_edge_length
    ...         (10, -10, 0, 0.1),
    ...         (10, 0, 0, 1e-2),
    ...         (0, 0, 0, 1e-2),
    ...     ]
    ...     model.add_tags()
    ...     model.add_physical_groups()
    >>> model.surface_tags
    [1]
    >>> model.physical_line_tags
    [6, 7, 8, 9]
    >>> model.physical_group_tags
    [10]

    """

    # https://gitlab.onelab.info/gmsh/gmsh/-/raw/master/api/gmsh.py

    name: str
    dim: int
    optimize_args: dict = field(default_factory=dict)
    point_constraints: list = field(default_factory=list)
    point_tags: list = field(default_factory=list)
    line_tags: list = field(default_factory=list)
    loop_tags: list = field(default_factory=list)
    surface_tags: list = field(default_factory=list)
    physical_line_tags: list = field(default_factory=list)
    physical_group_tags: list = field(default_factory=list)
    mesh_info: dict = field(default_factory=dict)
    _write_file: bool = True
    _was_entered: bool = False

    # TODO: Possible attributes worth adding, no particular order:
    # - boundary/boundaries, see gm.model.getBoundary()
    # - bounding_box, see gm.model.getBoundingBox()
    # - wrapper method for gm.model.getClosestPoint() ? maybe we need entities for that

    def __enter__(self):
        # See: <https://gitlab.onelab.info/gmsh/gmsh/-/issues/1142>
        gm.initialize(["-noenv"])  # Don't let gmsh mess with $PYTHONPATH and $PATH.
        gm.model.add(self.name)
        self._was_entered = True
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        gm.model.geo.synchronize()
        self.add_physical_groups()
        gm.model.mesh.generate(self.dim)
        # Populate some mesh info for later reference.
        node_tags, node_coords, _ = gm.model.mesh.getNodes()
        self.mesh_info["node_tags"] = node_tags
        self.mesh_info["node_coords"] = node_coords.reshape((node_tags.size, 3))
        element_types, element_tags, _ = gm.model.mesh.getElements()
        self.mesh_info["element_types"] = element_types
        self.mesh_info["element_tags"] = element_tags
        edge_tags, edge_orientations = gm.model.mesh.getAllEdges()
        self.mesh_info["edge_tags"] = edge_tags
        self.mesh_info["edge_orientations"] = edge_orientations
        tri_face_tags, tri_face_nodes = gm.model.mesh.getAllFaces(3)
        self.mesh_info["tri_face_tags"] = tri_face_tags
        self.mesh_info["tri_face_nodes"] = tri_face_nodes
        quad_face_tags, quad_face_nodes = gm.model.mesh.getAllFaces(4)
        self.mesh_info["quad_face_tags"] = quad_face_tags
        self.mesh_info["quad_face_nodes"] = quad_face_nodes
        if len(self.optimize_args) > 0:
            gm.model.mesh.optimize(**self.optimize_args)
        _log.info(
            f"Line tags (anticlockwise from bottom left): {self.physical_line_tags}"
        )
        _log.info(f"Surface tags (check order???): {self.physical_group_tags}")
        if self._write_file:
            gm.write(f"{self.name}.msh")
        gm.finalize()

    def add_tags(self):
        """Attempt to generate necessary tags (for rectangular meshes) automatically."""
        if self._was_entered:
            if self.dim == 3:
                raise _err.ModelContextError(
                    "cannot generate automatic tags for 3D model."
                )

            self.point_tags = [
                gm.model.geo.addPoint(*point) for point in self.point_constraints
            ]
            for i, point in enumerate(self.point_tags):
                self.line_tags.append(
                    gm.model.geo.addLine(
                        point, self.point_tags[(i + 1) % len(self.point_tags)]
                    )
                )
            self.loop_tags.append(gm.model.geo.addCurveLoop(self.line_tags))
            self.surface_tags.append(gm.model.geo.addPlaneSurface([self.loop_tags[-1]]))

        else:
            raise _err.ModelContextError(
                "cannot generate tags of uninitialized model. Have you used a `with` block?"
            )

    def add_physical_groups(self):
        """Add physical group tags that are used by external tools (e.g. Fluidity)."""
        if self._was_entered:
            if self.dim == 3:
                raise _err.ModelContextError(
                    "cannot generate automatic physical groups for 3D model."
                )
            # First arg to addPhysicalGroup() is the component dimension: 0D, 1D, 2D or 3D
            self.physical_line_tags = [
                gm.model.addPhysicalGroup(1, [line]) for line in self.line_tags
            ]
            self.physical_group_tags = [
                gm.model.addPhysicalGroup(2, [self.surface_tags[-1]])
            ]
        else:
            raise _err.ModelContextError(
                "cannot add physical groups to uninitialized model. Have you used a `with` block?"
            )


def rectangle(name, ref_axes, center, width, height, resolution, **kwargs):
    """Generate a rectangular (2D) mesh.

    Examples:

    >>> rect = rectangle(
    ...     "test_rect",
    ...     ("x", "z"),
    ...     center=(0, 0),
    ...     width=1,
    ...     height=1,
    ...     resolution={"global": 1e-2},
    ...     _write_file=False
    ... )
    >>> rect.dim
    2
    >>> rect.name
    'test_rect'
    >>> rect.line_tags
    [1, 2, 3, 4]
    >>> rect.loop_tags
    [1]
    >>> [p[-1] for p in rect.point_constraints]
    [0.01, 0.01, 0.01, 0.01]

    >>> rect = rectangle(
    ...     "test_rect",
    ...     ("x", "z"),
    ...     center=(0, 0),
    ...     width=1,
    ...     height=1,
    ...     resolution={"north": 1e-2, "south": 1e-3},
    ...     _write_file=False
    ... )
    >>> [p[-1] for p in rect.point_constraints]
    [0.001, 0.001, 0.01, 0.01]

    >>> rect = rectangle(
    ...     "test_rect",
    ...     ("x", "z"),
    ...     center=(0, 0),
    ...     width=1,
    ...     height=1,
    ...     resolution={"north-west": 1e-3, "south-east": 1e-2},
    ...     _write_file=False
    ... )
    >>> rect.point_constraints[1][-1]
    0.01
    >>> rect.point_constraints[3][-1]
    0.001
    >>> rect.point_constraints[0][-1] == rect.point_constraints[2][-1]
    True
    >>> rect.point_constraints[0][-1]
    0.0055

    """

    h, v = _geo.to_indices(*ref_axes)
    center_h, center_v = center
    point_constraints = np.zeros((4, 4))  # x, y, z, nearby_edge_length
    # TODO: Support "center" which should trigger creation of an additional
    # point_constraint that is not connected by lines but just anchors the central
    # resolution constraint (assuming this is possible in gmsh).
    _loc_map = {
        "global": range(4),
        "north": (2, 3),
        "south": (0, 1),
        "east": (1, 2),
        "west": (0, 3),
        "north-east": (2,),
        "north-west": (3,),
        "south-east": (1,),
        "south-west": (0,),
    }
    for i, p in enumerate(point_constraints):
        p[-1] = np.mean(list(resolution.values()))
        for k, res in resolution.items():
            if i in _loc_map[k]:
                p[-1] = res
        match i:
            case 0:
                p[h] = center_h - width / 2
                p[v] = center_v - height / 2
            case 1:
                p[h] = center_h + width / 2
                p[v] = center_v - height / 2
            case 2:
                p[h] = center_h + width / 2
                p[v] = center_v + height / 2
            case 3:
                p[h] = center_h - width / 2
                p[v] = center_v + height / 2

    with Model(name, 2, **kwargs) as model:
        model.point_constraints = point_constraints
        model.add_tags()
        model.add_physical_groups()
    return model


# def orthopiped():
#     """Generate an orthopiped (3D “box”) mesh."""
#     ...
#
#
# def annulus():
#     """Generate an annulus (2D) mesh."""
#     ...
#
#
# def annulus_part():
#     """Generate an annulus segment (2D) mesh."""
#     ...



r"""> PyDRex: Methods to calculate texture and strain diagnostics.

.. note::
    Calculations expect orientation matrices $a$ to represent passive
    (i.e. alias) rotations, which are defined in terms of the extrinsic ZXZ
    euler angles $ϕ, θ, φ$ as
    $$
    a = \begin{bmatrix}
            \cosφ\cosϕ - \cosθ\sinϕ\sinφ & \cosθ\cosϕ\sinφ + \cosφ\sinϕ & \sinφ\sinθ \cr
           -\sinφ\cosϕ - \cosθ\sinϕ\cosφ & \cosθ\cosϕ\cosφ - \sinφ\sinϕ & \cosφ\sinθ \cr
            \sinθ\sinϕ & -\sinθ\cosϕ & \cosθ
        \end{bmatrix}
    $$
    such that a[i, j] gives the direction cosine of the angle between the i-th
    grain axis and the j-th external axis (in the global Eulerian frame).

"""

import functools as ft

import numba as nb
import numpy as np
import scipy.linalg as la

from pydrex import geometry as _geo
from pydrex import logger as _log
from pydrex import stats as _stats
from pydrex import tensors as _tensors
from pydrex import utils as _utils

Pool, HAS_RAY = _utils.import_proc_pool()
if HAS_RAY:
    import ray

    from pydrex import distributed as _dstr


def elasticity_components(voigt_matrices):
    """Calculate elasticity decompositions for the given elasticity tensors.

    Args:
    - `voigt_matrices` (array) — the Nx6x6 Voigt matrix representations of the averaged
      elasticity tensors for a series of polycrystal textures

    Returns a dictionary with the following data series:
    - `'bulk_modulus'` — the computed bulk modulus for each Voigt matrix C
    - `'shear_modulus'` — the computed shear modulus for each Voigt matrix C
    - `'percent_anisotropy'` — for each Voigt matrix C, the total percentage of the
      norm of the elastic tensor ||C|| that deviates from the norm of the "closest"
      isotropic elasticity tensor
    - `'percent_hexagonal'` — for each C, the percentage of ||C|| attributed to
      hexagonally symmetric minerals
    - `'percent_tetragonal'` — for each C, the percentage of ||C|| attributed to
      tetragonally symmetric minerals
    - `'percent_orthorhombic'` — for each C, the percentage of ||C|| attributed to
      orthorhombically symmetric minerals
    - `'percent_monoclinic'` — for each C, the percentage of ||C|| attributed to
      monoclinically symmetric minerals
    - `'percent_triclinic'` — for each C, the percentage of ||C|| attributed to
      triclinically "symmetric" minerals (no mirror planes)
    - `'hexagonal_axis'` — for each C, the axis of hexagonal symmetry for the "closest"
      hexagonally symmetric approximation to C, a.k.a. the "transverse isotropy" axis

    .. note::
        Only 5 symmetry classes are relevant for elasticity decomposition,
        compared to the usual 6 used to describe crystal families.
        Crystals with cubic symmetry contribute to the isotropic elasticity tensor,
        because the lattice spacing is identical in all orthogonal directions.
        Note also that the trigonal crystal *system* is not a crystal family
        (it belongs to the hexagonal family).

    """
    n_matrices = len(voigt_matrices)
    out = {
        "bulk_modulus": np.empty(n_matrices),
        "shear_modulus": np.empty(n_matrices),
        "percent_anisotropy": np.empty(n_matrices),
        "percent_hexagonal": np.empty(n_matrices),
        "percent_tetragonal": np.empty(n_matrices),
        "percent_orthorhombic": np.empty(n_matrices),
        "percent_monoclinic": np.empty(n_matrices),
        "percent_triclinic": np.empty(n_matrices),
        "hexagonal_axis": np.empty((n_matrices, 3)),
    }
    for m, matrix in enumerate(voigt_matrices):
        voigt_matrix = _tensors.upper_tri_to_symmetric(matrix)
        stiffness_dilat, stiffness_deviat = _tensors.voigt_decompose(voigt_matrix)
        K = np.trace(stiffness_dilat) / 9  # Bulk modulus
        G = (np.trace(stiffness_deviat) - 3 * K) / 10  # Shear modulus
        out["bulk_modulus"][m] = K
        out["shear_modulus"][m] = G

        # Appendix A5, Browaeys & Chevrot, 2004.
        # The isotropic stiffness vector is independent of the coordinate system.
        isotropic_vector = np.hstack(
            (
                np.repeat(K + 4 * G / 3, 3),
                np.repeat(np.sqrt(2) * (K - 2 * G / 3), 3),
                np.repeat(2 * G, 3),
                np.repeat(0, 12),
            )
        )
        voigt_vector = _tensors.voigt_matrix_to_vector(voigt_matrix)
        out["percent_anisotropy"][m] = (
            la.norm(voigt_vector - isotropic_vector) / la.norm(voigt_vector) * 100
        )

        # Search for SCCS axes (orthogonal axes defined intrinsically by
        # eigenstrains of the elastic tensor).
        unpermuted_SCCS = np.empty((3, 3))
        eigv_dij = la.eigh(stiffness_dilat)[1]
        eigv_vij = la.eigh(stiffness_deviat)[1]
        # Averaging of eigenvectors to get the symmetry axes.
        for i in range(3):
            index_vij = 0  # [i(+1?)] of v_ij = eigvect that is nearest to the d_ij one.
            angle = 10  # Initialise to any number > 2π radians.
            for j in range(3):
                # Calculate angle between a pair of eigenvectors.
                # One eigenvector is taken from v_ij and one from d_ij.
                # Do not distinguish between vectors in opposite directions.
                dot_eigvects = np.dot(eigv_dij[:, i], eigv_vij[:, j])
                angle_eigvects = smallest_angle(eigv_dij[:, i], eigv_vij[:, j])
                if angle_eigvects < angle:
                    angle = angle_eigvects
                    # NOTE: Differences between ASPECT And original implementation:
                    # - In C++, std::copysign(1.0, 0.0) returns -1.
                    # - In Fortran 90, sign(1.0, 0.0) returns 1.
                    # - ASPECT implementation uses 'j' instead of 'j+1', doesn't seem to
                    # make a difference using the equivalent change for vec_SCSS = ...
                    index_vij = np.sign(dot_eigvects) * j if dot_eigvects != 0 else j
                    # index_vij = (
                    #     np.sign(dot_eigvects) * (j + 1)
                    #     if dot_eigvects != 0
                    #     else (j + 1)
                    # )

            # Add/subtract the nearest v_ij eigenvector multiplied by the signed index,
            # which effectively rotates the d_ij eigenvector, and then normalise it.
            vec_SCCS = (
                eigv_dij[:, i] + index_vij * eigv_vij[:, int(abs(index_vij))]
            ) / 2
            # vec_SCSS = (
            #     eigv_dij[:, i] + index_vij * eigv_vij[:, int(abs(index_vij) - 1)]
            # ) / 2
            vec_SCCS /= la.norm(vec_SCCS)
            unpermuted_SCCS[:, i] = vec_SCCS

        # Determine SCCS permutation that gives best hexagonal approximation.
        # This is achieved by minimising the "distance" between the voigt vector and its
        # projection onto the hexagonal symmetry subspace.
        # The elastic tensor is rotated into this system for decomposition.
        elastic_tensor = _tensors.voigt_to_elastic_tensor(voigt_matrix)
        distance = la.norm(voigt_vector)
        for i in range(3):
            permuted_SCCS = unpermuted_SCCS[:, [(i + j) % 3 for j in range(3)]]
            rotated_voigt_matrix = _tensors.elastic_tensor_to_voigt(
                _tensors.rotate(elastic_tensor, permuted_SCCS.transpose())
            )
            rotated_voigt_vector = _tensors.voigt_matrix_to_vector(rotated_voigt_matrix)
            mono_and_higher_vector = _tensors.mono_project(rotated_voigt_vector)
            tric_vector = rotated_voigt_vector - mono_and_higher_vector
            ortho_and_higher_vector = _tensors.ortho_project(mono_and_higher_vector)
            tetr_and_higher_vector = _tensors.tetr_project(ortho_and_higher_vector)
            hex_and_higher_vector = _tensors.hex_project(tetr_and_higher_vector)
            mono_vector = mono_and_higher_vector - ortho_and_higher_vector
            ortho_vector = ortho_and_higher_vector - tetr_and_higher_vector
            tetr_vector = tetr_and_higher_vector - hex_and_higher_vector
            hex_vector = hex_and_higher_vector - isotropic_vector

            δ = la.norm(rotated_voigt_vector - hex_and_higher_vector)
            if δ < distance:
                distance = δ

                percent = 100 / la.norm(voigt_vector)
                # print("\n", _tensors.voigt_vector_to_matrix(hex_vector))
                out["percent_triclinic"][m] = la.norm(tric_vector) * percent
                out["percent_monoclinic"][m] = la.norm(mono_vector) * percent
                out["percent_orthorhombic"][m] = la.norm(ortho_vector) * percent
                out["percent_tetragonal"][m] = la.norm(tetr_vector) * percent
                out["percent_hexagonal"][m] = la.norm(hex_vector) * percent
                # Last SCCS axis is always the hexagonal symmetry axis.
                out["hexagonal_axis"][m, ...] = permuted_SCCS[:, 2]
    return out


def bingham_average(orientations, axis="a"):
    """Compute Bingham average of orientation matrices.

    Returns the antipodally symmetric average orientation
    of the given crystallographic `axis`, or the a-axis by default.
    Valid axis specifiers are "a" for [100], "b" for [010] and "c" for [001].

    See also: [Watson (1966)](https://doi.org/10.1086%2F627211),
    [Mardia & Jupp, “Directional Statistics”](https://doi.org/10.1002/9780470316979).

    """
    match axis:
        case "a":
            row = 0
        case "b":
            row = 1
        case "c":
            row = 2
        case _:
            raise ValueError(f"axis must be 'a', 'b', or 'c', not {axis}")

    # https://courses.eas.ualberta.ca/eas421/lecturepages/orientation.html
    # Eigenvector corresponding to largest eigenvalue is the mean direction.
    # SciPy returns eigenvalues in ascending order (same order for vectors).
    # SciPy uses lower triangular entries by default, no need for all components.
    mean_vector = la.eigh(_stats._scatter_matrix(np.asarray(orientations), row))[1][
        :, -1
    ]
    return mean_vector / la.norm(mean_vector)


def finite_strain(deformation_gradient, driver="ev"):
    """Extract measures of finite strain from the deformation gradient.

    Extracts the largest principal strain value and the vector defining the axis of
    maximum extension (longest semiaxis of the finite strain ellipsoid) from the 3x3
    deformation gradient tensor.

    """
    # Get eigenvalues and eigenvectors of the left stretch (Cauchy-Green) tensor.
    B_λ, B_v = la.eigh(
        deformation_gradient @ deformation_gradient.transpose(),
        driver=driver,
    )
    # Stretch ratio is sqrt(λ) - 1, the (-1) is so that undeformed strain is 0 not 1.
    return np.sqrt(B_λ[-1]) - 1, B_v[:, -1]


def symmetry_pgr(orientations, axis="a"):
    r"""Compute texture symmetry eigenvalue diagnostics from grain orientation matrices.

    Compute Point, Girdle and Random symmetry diagnostics
    for ternary texture classification.
    Returns the tuple (P, G, R) where
    $$
    \begin{align\*}
    P &= (λ_{1} - λ_{2}) / N \cr
    G &= 2 (λ_{2} - λ_{3}) / N \cr
    R &= 3 λ_{3} / N
    \end{align\*}
    $$
    with $N$ the sum of the eigenvalues $λ_{1} ≥ λ_{2} ≥ λ_{3}$
    of the scatter (inertia) matrix.

    See e.g. [Vollmer (1990)](https://doi.org/10.1130/0016-7606(1990)102%3C0786:AAOEMT%3E2.3.CO;2).

    """
    match axis:
        case "a":
            row = 0
        case "b":
            row = 1
        case "c":
            row = 2
        case _:
            raise ValueError(f"axis must be 'a', 'b', or 'c', not {axis}")

    scatter = _stats._scatter_matrix(orientations, row)
    # SciPy uses lower triangular entries by default, no need for all components.
    eigvals_descending = la.eigvalsh(scatter)[::-1]
    sum_eigvals = np.sum(eigvals_descending)
    return (
        (eigvals_descending[0] - eigvals_descending[1]) / sum_eigvals,
        2 * (eigvals_descending[1] - eigvals_descending[2]) / sum_eigvals,
        3 * eigvals_descending[2] / sum_eigvals,
    )


def misorientation_indices(
    orientation_stack,
    system: _geo.LatticeSystem,
    bins=None,
    ncpus=None,
    pool=None,
):
    """Calculate M-indices for a series of polycrystal textures.

    Calculate M-index using `misorientation_index` for a series of texture snapshots.
    The `orientation_stack` is a NxMx3x3 array of orientations where N is the number of
    texture snapshots and M is the number of grains.

    Uses either Ray or the Python multiprocessing library to calculate texture indices
    for multiple snapshots simultaneously. The arguments `ncpus` and `pool` are only
    relevant to the latter option: if `ncpus` is `None` the number of CPU cores to use
    is chosen automatically based on the maximum number available to the Python
    interpreter, otherwise the specified number of cores is requested. Alternatively, an
    existing instance of `multiprocessing.Pool` can be provided.

    If Ray is installed, it will be automatically preferred. In this case, the number of
    parallel workers should be set upon initialisation of the Ray cluster (which can be
    distributed over the network).

    See `misorientation_index` for documentation of the remaining arguments.

    """
    if ncpus is not None and pool is not None:
        _log.warning("ignoring `ncpus` argument because a Pool was provided")
    m_indices = np.empty(len(orientation_stack))
    _run = ft.partial(
        misorientation_index,
        system=system,
        bins=bins,
    )
    if pool is None:
        if ncpus is None:
            ncpus = _utils.default_ncpus()
        with Pool(processes=ncpus) as pool:
            for i, out in enumerate(pool.imap(_run, orientation_stack)):
                m_indices[i] = out
    else:
        if HAS_RAY:
            m_indices = np.array(
                ray.get(
                    [
                        _dstr.misorientation_index.remote(
                            ray.put(a), system=system, bins=bins
                        )
                        for a in orientation_stack
                    ]
                )
            )
        else:
            for i, out in enumerate(pool.imap(_run, orientation_stack)):
                m_indices[i] = out
    return m_indices


def misorientation_index(orientations, system: _geo.LatticeSystem, bins=None):
    r"""Calculate M-index for polycrystal orientations.

    The `bins` argument is passed to `numpy.histogram`.
    If left as `None`, 1° bins will be used as recommended by the reference paper.
    The `symmetry` argument specifies the lattice system which determines intrinsic
    symmetry degeneracies and the maximum allowable misorientation angle.
    See `_geo.LatticeSystem` for supported systems.

    .. warning::
        This method must be able to allocate an array of shape
        $ \frac{N!}{2(N-2)!}× M^{2} $
        for N the length of `orientations` and M the number of symmetry operations for
        the given `system`.

    See [Skemer et al. (2005)](https://doi.org/10.1016/j.tecto.2005.08.023).

    """
    θmax = _stats._max_misorientation(system)
    misorientations_count, bin_edges = _stats.misorientation_hist(
        orientations, system, bins
    )
    # Compute counts of theoretical misorientation for an isotropic aggregate,
    # using the same bins (Skemer 2005 recommend 1° bins).
    misorientations_theory = np.array(
        [
            _stats.misorientations_random(bin_edges[i], bin_edges[i + 1], system)
            for i in range(len(misorientations_count))
        ]
    )
    # Equation 2 in Skemer 2005.
    return (θmax / (2 * len(misorientations_count))) * np.sum(
        np.abs(misorientations_theory - misorientations_count)
    )


def coaxial_index(orientations, axis1="b", axis2="a"):
    r"""Calculate coaxial “BA” index for a combination of two crystal axes.

    The BA index of [Mainprice et al. (2015)](https://doi.org/10.1144/SP409.8)
    is derived from the eigenvalue `symmetry` diagnostics and measures point vs girdle
    symmetry in the aggregate. $BA ∈ [0, 1]$ where $BA = 0$ corresponds to a perfect
    axial girdle texture and $BA = 1$ represents a point symmetry texture assuming that
    the random component $R$ is negligible. May be less susceptible to random
    fluctuations compared to the raw eigenvalue diagnostics.

    """
    P1, G1, _ = symmetry_pgr(orientations, axis=axis1)
    P2, G2, _ = symmetry_pgr(orientations, axis=axis2)
    return 0.5 * (2 - (P1 / (G1 + P1)) - (G2 / (G2 + P2)))


@nb.njit(fastmath=True)
def smallest_angle(vector, axis, plane=None):
    """Get smallest angle between a unit `vector` and a bidirectional `axis`.

    The axis is specified using either of its two parallel unit vectors.
    Optionally project the vector onto the `plane` (given by its unit normal)
    before calculating the angle.

    Examples:

    >>> from numpy import asarray as Ŋ
    >>> smallest_angle(Ŋ([1e0, 0e0, 0e0]), Ŋ([1e0, 0e0, 0e0]))
    0.0
    >>> smallest_angle(Ŋ([1e0, 0e0, 0e0]), Ŋ([0e0, 1e0, 0e0]))
    90.0
    >>> smallest_angle(Ŋ([1e0, 0e0, 0e0]), Ŋ([0e0, -1e0, 0e0]))
    90.0
    >>> smallest_angle(Ŋ([1e0, 0e0, 0e0]), Ŋ([np.sqrt(2), np.sqrt(2), 0e0]))
    45.0
    >>> smallest_angle(Ŋ([1e0, 0e0, 0e0]), Ŋ([-np.sqrt(2), np.sqrt(2), 0e0]))
    45.0

    """
    if plane is not None:
        _vector = vector - plane * np.dot(vector, plane)
    else:
        _vector = vector
    angle = np.rad2deg(
        np.arccos(
            np.clip(
                np.asarray(  # https://github.com/numba/numba/issues/3469
                    np.dot(_vector, axis)
                    / (np.linalg.norm(_vector) * np.linalg.norm(axis))
                ),
                -1,
                1,
            )
        )
    )
    if angle > 90:
        return 180 - angle
    return angle



"""> PyDRex: Mesh, configuration and supporting data Input/Output functions.

PyDRex can read/write three kinds of plain text files:
- PyDRex configuration files, which specify simulation parameters and initial conditions
- 'SCSV' files, CSV files with YAML frontmatter for (small) scientific datasets
- Mesh files via `meshio`, to set up final mineral positions in steady flows.

SCSV files are our custom CSV files with a YAML header. The header is used for data
attribution and metadata, as well as a column type spec. There is no official spec for
SCSV files at the moment but they should follow the format of existing  SCSV files in
the `data/` folder of the source repository. For supported cell types, see
`SCSV_TYPEMAP`.

"""

import collections as c
import csv
import functools as ft
import io
import os
import pathlib
import sys

if sys.version_info >= (3, 11):
    import tomllib
else:
    import tomli as tomllib

from importlib.resources import files

import h5py
import meshio
import numpy as np
import yaml
from tqdm import tqdm

from pydrex import core as _core
from pydrex import exceptions as _err
from pydrex import logger as _log
from pydrex import velocity as _velocity

DEFAULT_PARAMS = {
    "olivine_fraction": 1.0,
    "enstatite_fraction": 0.0,
    "stress_exponent": 1.5,
    "deformation_exponent": 3.5,
    "gbm_mobility": 125,
    "gbs_threshold": 0.3,
    "nucleation_efficiency": 5.0,
    "number_of_grains": 3500,
    "initial_olivine_fabric": "A",
}
"""Default simulation parameters."""

SCSV_TYPEMAP = {
    "string": str,
    "integer": int,
    "float": float,
    "boolean": bool,
    "complex": complex,
}
"""Mapping of supported SCSV field types to corresponding Python types."""

_SCSV_DEFAULT_TYPE = "string"
_SCSV_DEFAULT_FILL = ""


def extract_h5part(file, phase, fabric, n_grains, output):
    """Extract CPO data from Fluidity h5part file and save to canonical formats."""
    from pydrex.minerals import Mineral

    with h5py.File(file, "r") as f:
        for particle_id in f["Step#0/id"][:]:
            # Fluidity writes empty arrays to the particle data after they are deleted.
            # We need only the timesteps before deletion of this particle.
            steps = []
            for k in sorted(list(f.keys()), key=lambda s: int(s.lstrip("Step#"))):
                if f[f"{k}/x"].shape[0] >= particle_id:
                    steps.append(k)

            # Temporary data arrays.
            n_timesteps = len(steps)
            x = np.zeros(n_timesteps)
            y = np.zeros(n_timesteps)
            z = np.zeros(n_timesteps)
            orientations = np.empty((n_timesteps, n_grains, 3, 3))
            fractions = np.empty((n_timesteps, n_grains))

            strains = np.zeros(n_timesteps)
            for t, k in enumerate(
                tqdm(steps, desc=f"Extracting particle {particle_id}")
            ):
                # Extract particle position.
                x[t] = f[f"{k}/x"][particle_id - 1]
                y[t] = f[f"{k}/y"][particle_id - 1]
                z[t] = f[f"{k}/z"][particle_id - 1]

                # Extract CPO data.
                strains[t] = f[f"{k}/CPO_{n_grains * 10 + 1}"][particle_id - 1]
                vals = np.empty(n_grains * 10)
                for n in range(len(vals)):
                    vals[n] = f[f"{k}/CPO_{n+1}"][particle_id - 1]

                orientations[t] = np.array(
                    [
                        np.reshape(vals[n : n + 9], (3, 3))
                        for n in range(0, 9 * n_grains, 9)
                    ]
                )
                fractions[t] = vals[9 * n_grains :]

            _postfix = str(particle_id)
            _fractions = list(fractions)
            _orientations = list(orientations)
            mineral = Mineral(
                phase=phase,
                fabric=fabric,
                n_grains=n_grains,
                fractions_init=_fractions[0],
                orientations_init=_orientations[0],
            )
            mineral.fractions = _fractions
            mineral.orientations = _orientations
            mineral.save(output, postfix=_postfix)
            save_scsv(
                output[:-4] + f"_{_postfix}" + ".scsv",
                {
                    "delimiter": ",",
                    "missing": "-",
                    "fields": [
                        {
                            "name": "strain",
                            "type": "float",
                            "unit": "percent",
                            "fill": np.nan,
                        },
                        {
                            "name": "x",
                            "type": "float",
                            "unit": "m",
                            "fill": np.nan,
                        },
                        {
                            "name": "y",
                            "type": "float",
                            "unit": "m",
                            "fill": np.nan,
                        },
                        {
                            "name": "z",
                            "type": "float",
                            "unit": "m",
                            "fill": np.nan,
                        },
                    ],
                },
                [strains * 200, x, y, z],
            )


def read_scsv(file):
    """Read data from an SCSV file.

    Prints the YAML header section to output and returns a NamedTuple with columns of
    the csv data. See also `save_scsv`.

    """
    with open(resolve_path(file)) as fileref:
        yaml_lines = []
        csv_lines = []

        is_yaml = False
        for line in fileref:
            if line == "\n":  # Empty lines are skipped.
                continue
            if line == "---\n":
                if is_yaml:
                    is_yaml = False  # Second --- ends YAML section.
                    continue
                else:
                    is_yaml = True  # First --- begins YAML section.
                    continue

            if is_yaml:
                yaml_lines.append(line)
            else:
                csv_lines.append(line)

        metadata = yaml.safe_load(io.StringIO("".join(yaml_lines)))
        schema = metadata["schema"]
        if not _validate_scsv_schema(schema):
            raise _err.SCSVError(
                f"unable to parse SCSV schema from '{file}'."
                + " Check logging output for details."
            )
        reader = csv.reader(
            csv_lines, delimiter=schema["delimiter"], skipinitialspace=True
        )

        schema_colnames = [d["name"] for d in schema["fields"]]
        header_colnames = [s.strip() for s in next(reader)]
        if not schema_colnames == header_colnames:
            raise _err.SCSVError(
                f"schema field names must match column headers in '{file}'."
                + f" You've supplied schema fields\n{schema_colnames}"
                + f"\n with column headers\n{header_colnames}"
            )

        _log.info("reading SCSV file: %s", resolve_path(file))
        Columns = c.namedtuple("Columns", schema_colnames)
        # __dict__() and __slots__() of NamedTuples is empty :(
        # Set up some pretty printing instead to give a quick view of column names.
        Columns.__str__ = lambda self: f"Columns: {self._fields}"
        Columns._repr_pretty_ = lambda self, p, _: p.text(f"Columns: {self._fields}")
        # Also add some extra attributes to inspect the schema and yaml header.
        Columns._schema = schema
        Columns._metadata = (
            "".join(yaml_lines)
            .replace("# ", "")
            .replace("-\n", "")
            .replace("\n", " ")
            .rsplit("schema:", maxsplit=1)[0]  # Assumes comments are above the schema.
        )
        coltypes = [
            SCSV_TYPEMAP[d.get("type", _SCSV_DEFAULT_TYPE)] for d in schema["fields"]
        ]
        missingstr = schema["missing"]
        fillvals = [d.get("fill", _SCSV_DEFAULT_FILL) for d in schema["fields"]]
        return Columns._make(
            [
                tuple(
                    map(
                        ft.partial(
                            _parse_scsv_cell, f, missingstr=missingstr, fillval=fill
                        ),
                        x,
                    )
                )
                for f, fill, x in zip(
                    coltypes, fillvals, zip(*list(reader), strict=True), strict=True
                )
            ]
        )


def write_scsv_header(stream, schema, comments=None):
    """Write YAML header to an SCSV stream.

    Args:
    - `stream` — open output stream (e.g. file handle) where data should be written
    - `schema` — SCSV schema dictionary, with 'delimiter', 'missing' and 'fields' keys
    - `comments` (optional) — array of comments to be written above the schema, each on
      a new line with an '#' prefix

    See also `read_scsv`, `save_scsv`.

    """
    if not _validate_scsv_schema(schema):
        raise _err.SCSVError(
            "refusing to write invalid schema to stream."
            + " Check logging output for details."
        )

    stream.write("---" + os.linesep)
    if comments is not None:
        for comment in comments:
            stream.write("# " + comment + os.linesep)
    stream.write("schema:" + os.linesep)
    delimiter = schema["delimiter"]
    missing = schema["missing"]
    stream.write(f"  delimiter: '{delimiter}'{os.linesep}")
    stream.write(f"  missing: '{missing}'{os.linesep}")
    stream.write("  fields:" + os.linesep)

    for field in schema["fields"]:
        name = field["name"]
        kind = field.get("type", _SCSV_DEFAULT_TYPE)
        stream.write(f"    - name: {name}{os.linesep}")
        stream.write(f"      type: {kind}{os.linesep}")
        if "unit" in field:
            unit = field["unit"]
            stream.write(f"      unit: {unit}{os.linesep}")
        if "fill" in field:
            fill = field["fill"]
            stream.write(f"      fill: {fill}{os.linesep}")
    stream.write("---" + os.linesep)


def save_scsv(file, schema, data, **kwargs):
    """Save data to SCSV file.

    Args:
    - `file` — path to the file where the data should be written
    - `schema` — SCSV schema dictionary, with 'delimiter', 'missing' and 'fields' keys
    - `data` — data arrays (columns) of equal length

    Optional keyword arguments are passed to `write_scsv_header`. See also `read_scsv`.

    """
    path = resolve_path(file)
    n_rows = len(data[0])
    for col in data[1:]:
        if len(col) != n_rows:
            raise _err.SCSVError(
                "refusing to write data columns of unequal length to SCSV file"
            )

    _log.info("writing to SCSV file: %s", file)
    try:  # Check that the output is valid by attempting to parse.
        with open(path, mode="w") as stream:
            write_scsv_header(stream, schema, **kwargs)
            fills = [
                field.get("fill", _SCSV_DEFAULT_FILL) for field in schema["fields"]
            ]
            types = [
                SCSV_TYPEMAP[field.get("type", _SCSV_DEFAULT_TYPE)]
                for field in schema["fields"]
            ]
            names = [field["name"] for field in schema["fields"]]
            writer = csv.writer(
                stream, delimiter=schema["delimiter"], lineterminator=os.linesep
            )
            writer.writerow(names)

            # No need for strict=True here since column lengths were already checked.
            for col in zip(*data):
                row = []
                for i, (d, t, f) in enumerate(zip(col, types, fills, strict=True)):
                    try:
                        _parse_scsv_cell(
                            t, str(d), missingstr=schema["missing"], fillval=f
                        )
                    except ValueError:
                        raise _err.SCSVError(
                            f"invalid data for column '{names[i]}'."
                            + f" Cannot parse {d} as type '{t.__qualname__}'."
                        ) from None
                    if t == bool:
                        row.append(d)
                    elif t in (float, complex):
                        if np.isnan(d) and np.isnan(t(f)):
                            row.append(schema["missing"])
                        elif d == t(f):
                            row.append(schema["missing"])
                        else:
                            row.append(d)
                    elif t in (int, str) and d == t(f):
                        row.append(schema["missing"])
                    else:
                        row.append(d)
                writer.writerow(row)
    except ValueError:
        path.unlink(missing_ok=True)
        raise _err.SCSVError(
            "number of fields declared in schema does not match number of data columns."
            + f" Declared schema fields were {names}; got {len(data)} data columns"
        ) from None


def parse_config(path):
    """Parse a TOML file containing PyDRex configuration."""
    path = resolve_path(path)
    _log.info("parsing configuration file: %s", path)
    with open(path, "rb") as file:
        toml = tomllib.load(file)

    # Use provided name or set randomized default.
    toml["name"] = toml.get(
        "name", f"pydrex.{np.random.default_rng().integers(1,1e10)}"
    )

    _params = toml.get("parameters", {})
    for key, default in DEFAULT_PARAMS.items():
        _params[key] = _params.get(key, default)

    try:
        _input = toml["input"]
    except KeyError:
        raise _err.ConfigError(f"missing [input] section in '{path}'") from None
    if "timestep" not in _input and "paths" not in _input:
        raise _err.ConfigError(f"unspecified input timestep in '{path}'")

    _input["timestep"] = _input.get("timestep", np.nan)
    if not isinstance(_input["timestep"], float | int):
        raise _err.ConfigError(
            f"timestep must be float or int, not {type(input['timestep'])}"
        )

    _input["max_strain"] = _input.get("max_strain", np.inf)
    if not isinstance(_input["max_strain"], float | int):
        raise _err.ConfigError(
            f"timestep must be float or int, not {type(input['max_strain'])}"
        )

    # Input option 1: velocity gradient mesh + final particle locations.
    if "mesh" in _input:
        _input["mesh"] = meshio.read(resolve_path(_input["mesh"], path.parent))
        _input["locations_final"] = read_scsv(
            resolve_path(_input["locations_final"], path.parent)
        )
        if "velocity_gradient" in _input:
            _log.warning(
                "input mesh and velocity gradient callable are mutually exclusive;"
                + " ignoring velocity gradient callable"
            )
        if "locations_initial" in _input:
            _log.warning(
                "initial particle locations are not used for pathline interpolation"
                + " and will be ignored"
            )
        if "paths" in _input:
            _log.warning(
                "input mesh and input pathlines are mutually exclusive;"
                + " ignoring input pathlines"
            )
        _input["velocity_gradient"] = None
        _input["locations_initial"] = None
        _input["paths"] = None

    # Input option 2: velocity gradient callable + initial locations.
    elif "velocity_gradient" in _input:
        _velocity_gradient_func = getattr(_velocity, _input["velocity_gradient"][0])
        _input["velocity_gradient"] = _velocity_gradient_func(
            *_input["velocity_gradient"][1:]
        )
        _input["locations_initial"] = read_scsv(
            resolve_path(_input["locations_initial"], path.parent)
        )
        if "locations_final" in _input:
            _log.warning(
                "final particle locations are not used for forward advection"
                + " and will be ignored"
            )
        if "paths" in _input:
            _log.warning(
                "velocity gradient callable and input pathlines are mutually exclusive;"
                + " ignoring input pathlines"
            )
        _input["locations_final"] = None
        _input["paths"] = None
        _input["mesh"] = None

    # Input option 3: NPZ or SCSV files with pre-computed input pathlines.
    elif "paths" in _input:
        _input["paths"] = [
            np.load(resolve_path(p, path.parent)) for p in _input["paths"]
        ]
        if "locations_initial" in _input:
            _log.warning(
                "input pathlines and initial particle locations are mutually exclusive;"
                + " ignoring initial particle locations"
            )
        if "locations_final" in _input:
            _log.warning(
                "input pathlines and final particle locations are mutually exclusive;"
                + " ignoring final particle locations"
            )
        _input["locations_initial"] = None
        _input["locations_final"] = None
        _input["mesh"] = None
    else:
        _input["paths"] = None

    # Output fields are optional, default: most data output, least logging output.
    _output = toml.get("output", {})  # Defaults handled by _parse_output_options.
    if "directory" in _output:
        _output["directory"] = resolve_path(_output["directory"], path.parent)
    else:
        _output["directory"] = resolve_path(pathlib.Path.cwd())

    # Raw output means rotation matrices and grain volumes.
    _parse_output_options(_output, "raw_output")
    # Diagnostic output means texture diagnostics (strength, symmetry, mean angle).
    _parse_output_options(_output, "diagnostics")
    # Anisotropy output means hexagonal symmetry axis and ΔVp (%).
    _output["anisotropy"] = _output.get("anisotropy", True)

    # Optional SCSV or NPZ pathline outputs, not sensible if there are pathline inputs.
    if "paths" in _input and "paths" in _output:
        _log.warning(
            "input pathlines and output pathline filenames are mutually exclusive;"
            + " ignoring output pathline filenames"
        )
        _output["paths"] = None
    _output["paths"] = _output.get("paths", None)

    # Default logging level for all log files.
    _output["log_level"] = _output.get("log_level", "WARNING")

    # Only olivine and enstatite for now, so they must sum to 1.
    if _params["olivine_fraction"] + _params["enstatite_fraction"] != 1.0:
        raise _err.ConfigError(
            "olivine_fraction and enstatite_fraction must sum to 1."
            + f" You've provided olivine_fraction = {_params['olivine_fraction']} and"
            + f" enstatite_fraction = {_params['enstatite_fraction']}."
        )
    if _params["olivine_fraction"] != 1.0:
        _params["enstatite_fraction"] = 1 - _params["olivine_fraction"]
    elif _params["enstatite_fraction"] != 0.0:
        _params["olivine_fraction"] = 1 - _params["enstatite_fraction"]

    # Make sure initial olivine fabric is valid.
    try:
        _params["initial_olivine_fabric"] = getattr(
            _core.MineralFabric, "olivine_" + _params["initial_olivine_fabric"]
        )
    except AttributeError:
        raise _err.ConfigError(
            f"invalid initial olivine fabric: {_params['initial_olivine_fabric']}"
        ) from None
    return toml


def resolve_path(path, refdir=None):
    """Resolve relative paths and create parent directories if necessary.

    Relative paths are interpreted with respect to the current working directory,
    i.e. the directory from whith the current Python process was executed,
    unless a specific reference directory is provided with `refdir`.

    """
    cwd = pathlib.Path.cwd()
    if refdir is None:
        _path = cwd / path
    else:
        _path = refdir / path
    _path.parent.mkdir(parents=True, exist_ok=True)
    return _path.resolve()


def _parse_output_options(output_opts, level):
    try:
        output_opts[level] = [
            getattr(_core.MineralPhase, phase)
            for phase in output_opts.get(level, ["olivine", "enstatite"])
        ]
    except AttributeError:
        raise _err.ConfigError(
            f"unsupported mineral phase in {level} option.\n"
            + f" You supplied the value: {output_opts[level]}.\n"
            + " Check pydrex.core.MineralPhase for supported options."
        ) from None


def _validate_scsv_schema(schema):
    format_ok = (
        "delimiter" in schema
        and "missing" in schema
        and "fields" in schema
        and len(schema["fields"]) > 0
        and schema["delimiter"] != schema["missing"]
        and schema["delimiter"] not in schema["missing"]
    )
    if not format_ok:
        _log.error(
            "invalid format for SCSV schema: %s"
            + "\nMust contain: 'delimiter', 'missing', 'fields'"
            + "\nMust contain at least one field."
            + "\nMust contain compatible 'missing' and 'delimiter' values.",
            schema,
        )
        return False
    for field in schema["fields"]:
        if not field["name"].isidentifier():
            _log.error(
                "SCSV field name '%s' is not a valid Python identifier", field["name"]
            )
            return False
        if field.get("type", _SCSV_DEFAULT_TYPE) not in SCSV_TYPEMAP.keys():
            _log.error("unsupported SCSV field type: '%s'", field["type"])
            return False
        if (
            field.get("type", _SCSV_DEFAULT_TYPE) not in (_SCSV_DEFAULT_TYPE, "boolean")
            and "fill" not in field
        ):
            _log.error("SCSV field of type '%s' requires a fill value", field["type"])
            return False
    return True


def _parse_scsv_bool(x):
    """Parse boolean from string, for SCSV files."""
    return str(x).lower() in ("yes", "true", "t", "1")


def _parse_scsv_cell(func, data, missingstr=None, fillval=None):
    if data.strip() == missingstr:
        if fillval == "NaN":
            return func(np.nan)
        return func(fillval)
    elif func == bool:
        return _parse_scsv_bool(data)
    return func(data.strip())


def stringify(s):
    """Return a cleaned version of a string for use in filenames, etc."""
    return "".join(filter(lambda c: str.isidentifier(c) or str.isdecimal(c), str(s)))


def data(directory):
    """Get resolved path to a pydrex data directory."""
    resources = files("pydrex.data")
    if (resources / directory).is_dir():
        return resolve_path(resources / directory)
    else:
        raise NotADirectoryError(f"{resources / directory} is not a directory")



"""> PyDRex: tests for CPO stability in 2D vortex and Stokes cell flows."""

import functools as ft
import sys

import numpy as np
import pytest

from pydrex import core as _core
from pydrex import diagnostics as _diagnostics
from pydrex import io as _io
from pydrex import logger as _log
from pydrex import minerals as _minerals
from pydrex import pathlines as _path
from pydrex import utils as _utils
from pydrex import velocity as _velocity
from pydrex import visualisation as _vis

Pool, HAS_RAY = _utils.import_proc_pool()

# Subdirectory of `outdir` used to store outputs from these tests.
SUBDIR = "2d_vortex"


class TestCellOlivineA:
    """Tests for A-type olivine polycrystals in a 2D Stokes cell."""

    class_id = "cell_olivineA"
    _ensemble_n_grains = [100, 500, 1000, 5000, 10000]

    @classmethod
    def _make_ensemble_figure(cls, outdir):
        # Create the combined figure from outputs of the parametrized ensemble test.
        data = []
        out_basepath = f"{outdir}/{SUBDIR}/{cls.class_id}"
        try:
            for n_grains in cls._ensemble_n_grains:
                data.append(np.load(f"{out_basepath}_xz_ensemble_N{n_grains}_data.npz"))
        except FileNotFoundError:
            _log.debug(
                "skipping visualisation of 2D cell ensemble results (missing datafiles)"
            )
            return

        fig = _vis.figure()
        mosaic = fig.subplot_mosaic([["a)"], ["b)"]], sharex=True)
        fig, mosaic["a)"], colors = _vis.alignment(
            mosaic["a)"],
            np.asarray([d["strains"] for d in data]),
            np.asarray([d["angles_mean"] for d in data]),
            ("s", "o", "v", "*", "."),
            list(map(str, cls._ensemble_n_grains)),
            err=np.asarray([d["angles_err"] for d in data]),
        )
        fig, mosaic["b)"], colors = _vis.alignment(
            mosaic["b)"],
            np.asarray([d["strains"] for d in data]),
            np.asarray([d["max_sizes_mean"] for d in data]),
            ("s", "o", "v", "*", "."),
            list(map(str, cls._ensemble_n_grains)),
            err=np.asarray([d["max_sizes_err"] for d in data]),
            θ_max=4,
        )
        mosaic["b)"].set_ylabel(r"$\log_{10}(\overline{S}_{\mathrm{max}})$")

        mosaic["a)"].label_outer()
        mosaic["b)"].get_legend().remove()
        fig.savefig(
            _io.resolve_path(
                f"{outdir}/{SUBDIR}/{cls.class_id}_xz_ensemble_combined.pdf"
            )
        )

    @classmethod
    def run(
        cls,
        params,
        final_location,
        get_velocity,
        get_velocity_gradient,
        min_coords,
        max_coords,
        max_strain,
        seed=None,
    ):
        """Run 2D Stokes cell A-type olivine simulation."""
        mineral = _minerals.Mineral(
            phase=_core.MineralPhase.olivine,
            fabric=_core.MineralFabric.olivine_A,
            regime=_core.DeformationRegime.dislocation,
            n_grains=params["number_of_grains"],
            seed=seed,
        )
        deformation_gradient = np.eye(3)

        timestamps, get_position = _path.get_pathline(
            final_location,
            get_velocity,
            get_velocity_gradient,
            min_coords,
            max_coords,
            max_strain,
            regular_steps=int(max_strain * 10),
        )
        positions = [get_position(t) for t in timestamps]
        velocity_gradients = [
            get_velocity_gradient(np.nan, np.asarray(x)) for x in positions
        ]
        strains = np.empty_like(timestamps)
        strains[0] = 0
        for t, time in enumerate(timestamps[:-1], start=1):
            strains[t] = strains[t - 1] + (
                _utils.strain_increment(timestamps[t] - time, velocity_gradients[t])
            )
            _log.info("step %d/%d (ε = %.2f)", t, len(timestamps) - 1, strains[t])

            deformation_gradient = mineral.update_orientations(
                params,
                deformation_gradient,
                get_velocity_gradient,
                pathline=(time, timestamps[t], get_position),
            )
        return timestamps, positions, strains, mineral, deformation_gradient

    @pytest.mark.big
    def test_xz_10k(self, outdir, seed):
        """Run 2D cell test with 10000 grains (~14GiB RAM requirement)."""
        self.test_xz(outdir, seed, 10000)

    @pytest.mark.skipif(sys.platform == "win32", reason="Unable to allocate memory")
    @pytest.mark.parametrize("n_grains", [100, 500, 1000, 5000])
    def test_xz(self, outdir, seed, n_grains):
        """Test to check that 5000 grains is "enough" to resolve transient features."""
        if outdir is not None:
            out_basepath = f"{outdir}/{SUBDIR}/{self.class_id}_xz_N{n_grains}"

        params = _io.DEFAULT_PARAMS
        params["number_of_grains"] = n_grains
        get_velocity, get_velocity_gradient = _velocity.cell_2d("X", "Z", 1)

        timestamps, positions, strains, mineral, deformation_gradient = self.run(
            params,
            np.asarray([0.5, 0.0, -0.75]),
            get_velocity,
            get_velocity_gradient,
            np.asarray([-1, 0, -1]),
            np.asarray([1, 0, 1]),
            7,
            seed=seed,
        )
        angles = [
            _diagnostics.smallest_angle(
                _diagnostics.bingham_average(a, axis="a"), get_velocity(np.nan, x)
            )
            for a, x in zip(mineral.orientations, positions, strict=True)
        ]
        if outdir is not None:
            # First figure with the domain and pathline.
            fig_path, ax_path, q, s = _vis.pathline_box2d(
                None,
                get_velocity,
                "XZ",
                strains,
                positions,
                ".",
                [-1, -1],
                [1, 1],
                [20, 20],
                cmap="cmc.batlow_r",
                tick_formatter=lambda x, pos: str(x),
                aspect="equal",
                alpha=1,
            )
            fig_path.colorbar(s, ax=ax_path, aspect=25, label="Strain (ε)")
            fig_path.savefig(_io.resolve_path(f"{out_basepath}_path.pdf"))
            # Second figure with the angles and grain sizes at every 10 strain values.
            fig = _vis.figure()
            axθ = fig.add_subplot(2, 1, 1)
            fig, axθ, colors = _vis.alignment(
                axθ,
                strains,
                angles,
                (".",),
                (None,),
                colors=[strains],
                cmaps=["cmc.batlow_r"],
            )
            ax_sizes = fig.add_subplot(2, 1, 2, sharex=axθ)
            fig, ax_sizes, parts = _vis.grainsizes(
                ax_sizes, strains[::10], mineral.fractions[::10]
            )
            axθ.label_outer()
            fig.savefig(_io.resolve_path(f"{out_basepath}.pdf"))

        # Some checks for when we should have "enough" grains.
        # Based on empirical model outputs, it seems like the dip at ε ≈ 3.75 is the
        # least sensitive feature to the random state (seed) so we will use that.
        if n_grains >= 5000:
            # Can we resolve the temporary alignment to below 20° at ε ≈ 3.75?
            mean_θ_in_dip = np.mean(angles[34:43])
            assert mean_θ_in_dip < 12, mean_θ_in_dip
            # Can we resolve corresponding dip in max grain size (normalized, log_10)?
            mean_size_in_dip = np.log10(
                np.mean([np.max(f) for f in mineral.fractions[34:43]]) * n_grains
            )
            assert 2 < mean_size_in_dip < 3, mean_size_in_dip
            # Can we resolve subsequent peak in max grain size (normalized, log_10)?
            max_size_post_dip = np.log10(
                np.max([np.max(f) for f in mineral.fractions[43:]]) * n_grains
            )
            assert max_size_post_dip > 3, max_size_post_dip

    @pytest.mark.slow
    @pytest.mark.parametrize("n_grains", [100, 500, 1000, 5000, 10000])
    def test_xz_ensemble(self, outdir, seeds_nearX45, ncpus, n_grains):
        """Test to demonstrate stability of the dip at ε ≈ 3.75 for 5000+ grains."""
        _seeds = seeds_nearX45
        n_seeds = len(_seeds)
        if outdir is not None:
            out_basepath = f"{outdir}/{SUBDIR}/{self.class_id}_xz_ensemble_N{n_grains}"

        params = _io.DEFAULT_PARAMS
        params["number_of_grains"] = n_grains
        get_velocity, get_velocity_gradient = _velocity.cell_2d("X", "Z", 1)

        _run = ft.partial(
            self.run,
            params,
            np.asarray([0.5, 0.0, -0.75]),
            get_velocity,
            get_velocity_gradient,
            np.asarray([-1, 0, -1]),
            np.asarray([1, 0, 1]),
            7,
        )
        angles = np.empty((n_seeds, 70))
        max_sizes = np.empty_like(angles)
        with Pool(processes=ncpus) as pool:
            for s, out in enumerate(pool.imap_unordered(_run, _seeds)):
                timestamps, positions, strains, mineral, deformation_gradient = out
                angles[s] = [
                    _diagnostics.smallest_angle(
                        _diagnostics.bingham_average(a, axis="a"),
                        get_velocity(np.nan, x),
                    )
                    for a, x in zip(mineral.orientations, positions, strict=True)
                ]
                max_sizes[s] = np.log10(np.max(mineral.fractions, axis=1) * n_grains)

        if outdir is not None:
            # Figure with the angles and max grain sizes (ensemble averages).
            fig = _vis.figure()
            axθ = fig.add_subplot(2, 1, 1)
            angles_mean = np.mean(angles, axis=0)
            angles_err = np.std(angles, axis=0)
            fig, axθ, colors = _vis.alignment(
                axθ,
                strains,
                angles_mean,
                (".",),
                (None,),
                err=angles_err,
            )
            ax_maxsize = fig.add_subplot(2, 1, 2, sharex=axθ)
            ax_maxsize.set_ylabel(r"$\log_{10}(\overline{S}_{\mathrm{max}})$")
            max_sizes_mean = np.mean(max_sizes, axis=0)
            ax_maxsize.plot(strains, max_sizes_mean, color=colors[0])
            max_sizes_err = np.std(max_sizes, axis=0)
            ax_maxsize.fill_between(
                strains,
                max_sizes_mean - max_sizes_err,
                max_sizes_mean + max_sizes_err,
                alpha=0.22,
                color=colors[0],
            )
            axθ.label_outer()
            fig.savefig(_io.resolve_path(f"{out_basepath}.pdf"))
            np.savez(
                _io.resolve_path(f"{out_basepath}_data.npz"),
                strains=strains,
                max_sizes_mean=max_sizes_mean,
                max_sizes_err=max_sizes_err,
                angles_mean=angles_mean,
                angles_err=angles_err,
            )



#!/usr/bin/env python3
###############################################################################
# Python 3 Version of D-Rex
###############################################################################
import argparse
from time import perf_counter
from warnings import catch_warnings, simplefilter

import numpy as np
from DRexParam import (
    S0,
    S0_ens,
    Xol,
    checkpoint,
    chi,
    gridCoords,
    gridMax,
    gridMin,
    gridNodes,
    lamb,
    mob,
    name,
    size,
    stressexp,
    tau,
    tau_ens,
)
from numba import jit
from numpy.linalg import eigh, eigvalsh, norm
from scipy.integrate import solve_ivp
from scipy.interpolate import NearestNDInterpolator
from scipy.spatial.transform import Rotation as R
from vtk import vtkXMLPUnstructuredGridReader, vtkXMLUnstructuredGridReader

###############################################################################
#
# symFromUpper    : Generates a symmetric array from the upper triangle part of
#                   another array
# formElasticTens : Yields the 4th-order elastic tensor equivalent to the 6x6
#                   Voigt matrix
# formVoigtMat    : Yields the 6x6 Voigt matrix equivalent to the 4th-order
#                   elastic tensor
# formVoigtVec    : Yields the 21-component Voigt vector equivalent to the 6x6
#                   Voigt matrix
# rotateTens      : Rotates a 4th-order tensor
# trIsoProj       : Transverse isotropy projector
# scca            : Forms symmetric cartesian coordinate system
# calcAzimuth     : Calculates the azimuthal fast direction in a horizontal
#                   plane
# decsym          : Decomposition into transverse isotropy symmetry
# interpVel       : Interpolates the velocity vector at a given point
# interpVelGrad   : Interpolates the velocity gradient tensor at a given point
# fseDecomp       : Rotates a matrix line based on finite strain ellipsoid
# deriv           : Calculation of the rotation vector and slip rate
# strain          : Calculation of strain along pathlines
# voigt           : Calculates elastic tensor cav_{ijkl} for olivine
# pipar           : Calculates GOL parameter at grid point
# isacalc         : Calculates ISA Orientation at grid point
# pathline        : Determines a pathline given a position and a velocity field
# isInside        : Checks if a point lies within the numerical domain
# DRex            : Main function
# main            : Init function
#
###############################################################################


# Generates a symmetric array from the upper triangle part of another array
def symFromUpper(arr):
    return np.triu(arr) + np.transpose(np.triu(arr)) - np.diag(np.diag(arr))


# Yields the 4th-order elastic tensor equivalent to the 6x6 Voigt matrix
def formElasticTens(mat):
    # Equation 2.1 Browaeys and Chevrot (2004)
    tens = np.empty((3, 3, 3, 3))
    for p in range(3):
        for q in range(3):
            deltaPQ = 1 if p == q else 0
            i = (p + 1) * deltaPQ + (1 - deltaPQ) * (7 - p - q) - 1
            for r in range(3):
                for s in range(3):
                    deltaRS = 1 if r == s else 0
                    j = (r + 1) * deltaRS + (1 - deltaRS) * (7 - r - s) - 1
                    tens[p, q, r, s] = mat[i, j]
    return tens


# Yields the 6x6 Voigt matrix equivalent to the 4th-order elastic tensor
def formVoigtMat(tens):
    # Equation 2.1 Browaeys and Chevrot (2004)
    # Components c_{ijkl} which map to C_{IJ} and C_{JI} (Table 1) are averaged
    mat = np.zeros((6, 6))
    matInd = np.zeros((6, 6))
    for p in range(3):
        for q in range(3):
            deltaPQ = 1 if p == q else 0
            i = (p + 1) * deltaPQ + (1 - deltaPQ) * (7 - p - q) - 1
            for r in range(3):
                for s in range(3):
                    deltaRS = 1 if r == s else 0
                    j = (r + 1) * deltaRS + (1 - deltaRS) * (7 - r - s) - 1
                    mat[i, j] += tens[p, q, r, s]
                    matInd[i, j] += 1
    mat /= matInd
    return (mat + mat.transpose()) / 2


# Yields the 21-component Voigt vector equivalent to the 6x6 Voigt matrix
def formVoigtVec(mat):
    # Equation 2.2 Browaeys and Chevrot (2004)
    vec = np.zeros(21)
    for i in range(3):
        vec[i] = mat[i, i]
        vec[i + 3] = np.sqrt(2) * mat[(i + 1) % 3, (i + 2) % 3]
        vec[i + 6] = 2 * mat[i + 3, i + 3]
        vec[i + 9] = 2 * mat[i, i + 3]
        vec[i + 12] = 2 * mat[(i + 2) % 3, i + 3]
        vec[i + 15] = 2 * mat[(i + 1) % 3, i + 3]
        vec[i + 18] = 2 * np.sqrt(2) * mat[(i + 1) % 3 + 3, (i + 2) % 3 + 3]
    return vec


# Rotates a 4th-order tensor
@jit(nopython=True)
def rotateTens(tens, rot):
    rotTens = np.zeros((3, 3, 3, 3))
    for i in range(3):
        for j in range(3):
            for k in range(3):
                for L in range(3):
                    for a in range(3):
                        for b in range(3):
                            for c in range(3):
                                for d in range(3):
                                    rotTens[i, j, k, L] += (
                                        rot[i, a]
                                        * rot[j, b]
                                        * rot[k, c]
                                        * rot[L, d]
                                        * tens[a, b, c, d]
                                    )
    return rotTens


# Transverse isotropy projector
def trIsoProj(x):
    # Appendix A4 Browaeys and Chevrot (2004)
    # An elastic medium with hexagonal symmetry is equivalent to a transversely
    # isotropic medium
    y = np.zeros(21)
    y[0] = y[1] = 3 / 8 * (x[0] + x[1]) + x[5] / 4 / np.sqrt(2) + x[8] / 4
    y[2] = x[2]
    y[3] = y[4] = (x[3] + x[4]) / 2
    y[5] = (x[0] + x[1]) / 4 / np.sqrt(2) + 3 / 4 * x[5] - x[8] / 2 / np.sqrt(2)
    y[6] = y[7] = (x[6] + x[7]) / 2
    y[8] = (x[0] + x[1]) / 4 - x[5] / 2 / np.sqrt(2) + x[8] / 2
    return norm(x - y, ord=2)


# Forms symmetric cartesian coordinate system
def scca(voigtMat):
    # Equation 3.4 Browaeys and Chevrot (2004)
    dilatStifTens = np.empty((3, 3))
    for i in range(3):
        dilatStifTens[i, i] = voigtMat[:3, i].sum()
    dilatStifTens[0, 1] = dilatStifTens[1, 0] = voigtMat[:3, 5].sum()
    dilatStifTens[0, 2] = dilatStifTens[2, 0] = voigtMat[:3, 4].sum()
    dilatStifTens[1, 2] = dilatStifTens[2, 1] = voigtMat[:3, 3].sum()
    # Equation 3.5 Browaeys and Chevrot (2004)
    voigtStifTens = np.empty((3, 3))
    voigtStifTens[0, 0] = voigtMat[0, 0] + voigtMat[4, 4] + voigtMat[5, 5]
    voigtStifTens[1, 1] = voigtMat[1, 1] + voigtMat[3, 3] + voigtMat[5, 5]
    voigtStifTens[2, 2] = voigtMat[2, 2] + voigtMat[3, 3] + voigtMat[4, 4]
    voigtStifTens[0, 1] = voigtMat[0, 5] + voigtMat[1, 5] + voigtMat[3, 4]
    voigtStifTens[0, 2] = voigtMat[0, 4] + voigtMat[2, 4] + voigtMat[3, 5]
    voigtStifTens[1, 2] = voigtMat[1, 3] + voigtMat[2, 3] + voigtMat[4, 5]
    voigtStifTens = symFromUpper(voigtStifTens)
    # Appendix A5 Browaeys and Chevrot (2004)
    K = np.trace(dilatStifTens) / 9  # Bulk modulus
    G = (np.trace(voigtStifTens) - 3 * K) / 10  # Shear modulus
    isoVec = np.hstack(
        (
            np.repeat(K + 4 * G / 3, 3),
            np.repeat(np.sqrt(2) * (K - 2 * G / 3), 3),
            np.repeat(2 * G, 3),
            np.repeat(0, 12),
        )
    )
    voigtVec = formVoigtVec(voigtMat)
    anisotropy = norm(voigtVec - isoVec)
    # Section 3.2 Browaeys and Chevrot (2004)
    eigVecDST = eigh(dilatStifTens)[1][:, ::-1]
    eigVecVST = eigh(voigtStifTens)[1][:, ::-1]
    # Search for SCCA directions
    for i in range(3):
        ndvc = 0
        advc = 10
        for j in range(3):
            sdv = np.clip(np.dot(eigVecDST[:, i], eigVecVST[:, j]), -1, 1)
            adv = np.arccos(abs(sdv))
            if adv < advc:
                ndvc = int((j + 1) * np.sign(sdv)) if sdv != 0 else j + 1
                advc = adv
        eigVecDST[:, i] = (eigVecDST[:, i] + ndvc * eigVecVST[:, abs(ndvc) - 1]) / 2
        eigVecDST[:, i] /= norm(eigVecDST[:, i], ord=2)
    # Higher symmetry axis
    elasTens = formElasticTens(voigtMat)
    normVec = norm(voigtVec)
    for i in range(3):
        dev = trIsoProj(
            formVoigtVec(
                formVoigtMat(
                    rotateTens(
                        elasTens,
                        eigVecDST[:, [(i + j) % 3 for j in range(3)]].transpose(),
                    )
                )
            )
        )
        if dev < normVec:
            normVec = dev
            ndvc = i + 1
    # Rotate in SCCA
    scc = eigVecDST[:, [(abs(ndvc) - 1 + i) % 3 for i in range(3)]].transpose()
    voigtVec = formVoigtVec(formVoigtMat(rotateTens(elasTens, scc)))
    return anisotropy, scc, voigtVec


# Calculates the azimuthal fast direction in a horizontal plane
def calcAzimuth(voigtMat, azi=0, inc=-np.pi / 2):
    # Code is adapted from the MATLAB Seismic Anisotropy Toolkit
    # https://github.com/andreww/MSAT/blob/master/msat/MS_phasevels.m
    # Create the cartesian vector
    Xr = np.array([np.cos(azi) * np.cos(inc), -np.sin(azi) * np.cos(inc), np.sin(inc)])
    Xr /= norm(Xr)
    # Compute Eigenvector
    gamma = np.array(
        [
            [Xr[0], 0, 0, 0, Xr[2], Xr[1]],
            [0, Xr[1], 0, Xr[2], 0, Xr[0]],
            [0, 0, Xr[2], Xr[1], Xr[0], 0],
        ]
    )
    # Form the 3x3 Christoffel tensor
    T = np.dot(np.dot(gamma, voigtMat), gamma.transpose())
    S1 = eigh(T)[1][:, 1]
    # Calculate projection onto propagation plane
    S1P = np.cross(Xr, np.cross(Xr, S1))
    # Rotate into YZ plane to calculate angles
    rot1 = np.array(
        [[np.cos(azi), np.sin(azi), 0], [-np.sin(azi), np.cos(azi), 0], [0, 0, 1]]
    )
    rot2 = np.array(
        [[np.cos(inc), 0, -np.sin(inc)], [0, 1, 0], [np.sin(inc), 0, np.cos(inc)]]
    )
    VR = np.dot(np.dot(S1P, rot1), rot2)
    return -np.arctan2(VR[1], VR[2])


# Decomposition into transverse isotropy symmetry
def decsym(Sav):
    voigtMat = symFromUpper(Sav)
    # Original approach
    anisotropy, scc, voigtVec = scca(voigtMat)
    perc = (anisotropy - trIsoProj(voigtVec)) / norm(voigtVec) * 100
    tiAxis = scc[2, :]
    tiAxis /= norm(tiAxis)
    inclti = np.arcsin(tiAxis[2])
    # New approach
    radani = (
        (voigtMat[0, 0] + voigtMat[1, 1]) / 8 - voigtMat[0, 1] / 4 + voigtMat[5, 5] / 2
    ) / ((voigtMat[3, 3] + voigtMat[4, 4]) / 2)
    percani = np.sqrt(
        ((voigtMat[4, 4] - voigtMat[3, 3]) / 2) ** 2 + voigtMat[4, 3] ** 2
    )
    azimuth = calcAzimuth(voigtMat)
    return perc, inclti, radani, percani, azimuth


# Interpolates the velocity vector at a given point
def interpVel(pointCoords, dictGlobals):
    iVelX, iVelZ = dictGlobals["iVelX"], dictGlobals["iVelZ"]
    if len(pointCoords) == 3:
        iVelY = dictGlobals["iVelY"]
    pointVel = np.array(
        [
            iVelX(*pointCoords),
            (iVelY(*pointCoords) if len(pointCoords) == 3 else np.nan),
            -iVelZ(*pointCoords),
        ]
    )
    return pointVel[~np.isnan(pointVel)]


# Interpolates the velocity gradient tensor at a given point
def interpVelGrad(pointCoords, dictGlobals, ivpIter=False):
    iLxx, iLxz, iLzx, iLzz = (
        dictGlobals["iLxx"],
        dictGlobals["iLxz"],
        dictGlobals["iLzx"],
        dictGlobals["iLzz"],
    )
    if len(pointCoords) == 2:
        L = np.array(
            [
                [iLxx(*pointCoords), 0, -iLxz(*pointCoords)],
                [0, 0, 0],
                [-iLzx(*pointCoords), 0, iLzz(*pointCoords)],
            ]
        )
    else:
        iLxy, iLyx, iLyy, iLyz, iLzy = (
            dictGlobals["iLxy"],
            dictGlobals["iLyx"],
            dictGlobals["iLyy"],
            dictGlobals["iLyz"],
            dictGlobals["iLzy"],
        )
        L = np.array(
            [
                [iLxx(*pointCoords), iLxy(*pointCoords), -iLxz(*pointCoords)],
                [iLyx(*pointCoords), iLyy(*pointCoords), -iLyz(*pointCoords)],
                [-iLzx(*pointCoords), -iLzy(*pointCoords), iLzz(*pointCoords)],
            ]
        )
    assert abs(np.trace(L)) < 1e-15
    if ivpIter:
        return L
    e = (L + np.transpose(L)) / 2
    epsnot = abs(eigvalsh(e)).max()
    return L, epsnot


# Rotation of a matrix line based on finite strain ellipsoid
@jit(nopython=True)
def fseDecomp(fse, ex, alt):
    # Left-strech tensor for finite-strain ellipsoid calculation
    leftStretchTens = np.dot(fse, np.transpose(fse))
    eigval, eigvect = eigh(leftStretchTens)
    rot = np.zeros(3)  # Induced spin
    for i, eigv in enumerate(eigval):
        otherEig = eigval[np.array([(i + 1) % 3, (i + 2) % 3])]
        if (np.abs((otherEig - eigv) / eigv) < 5e-2).any():
            return rot
    # Equation 34 Ribe (1992)
    H = np.empty(3)
    for i in range(3):
        j, k = (i + 1) % 3, (i + 2) % 3
        H[i] = (
            eigval[i]
            * (eigval[j] - eigval[k])
            / (eigval[i] - eigval[j])
            / (eigval[i] - eigval[k])
        )
    # Equation 33 Ribe (1992) | alt -> epsilon, ex -> E, eigvect -> a.T
    for i in range(3):
        for n in range(3):
            m = (n + 2) % 3
            for j in range(3):
                for k in range(3):
                    rot[i] += (
                        H[n]
                        * ex[j, k]
                        * (
                            eigvect[j, n] * eigvect[k, (n + 1) % 3] * eigvect[i, m]
                            + eigvect[j, m]
                            * np.dot(
                                np.ascontiguousarray(alt[i, :, k]),
                                np.ascontiguousarray(eigvect[:, m]),
                            )
                        )
                    )
    return rot


# Calculation of the rotation vector and slip rate
@jit(nopython=True, fastmath=True)
def deriv(lx, ex, acsi, acsi_ens, fse, odfi, odfi_ens, alpha):
    alt = np.zeros((3, 3, 3))  # Levi-Civita symbol
    for i in range(3):
        alt[i % 3, (i + 1) % 3, (i + 2) % 3] = 1
        alt[i % 3, (i + 2) % 3, (i + 1) % 3] = -1
    g, g_ens = np.empty((3, 3)), np.empty((3, 3))
    rt, rt_ens = np.zeros(size), np.empty(size)
    dotacs, dotacs_ens = np.zeros((size, 3, 3)), np.zeros((size, 3, 3))
    if alpha == 0:
        rotIni = fseDecomp(fse, ex, alt)
    else:
        rot, rot_ens = np.empty(3), np.empty(3)
    gam = np.empty(4)
    # Plastic deformation & Dynamic recrystallization
    # Update olivine fabric type in the loop?
    for i in range(size):
        # Calculate invariants for the four slip systems of olivine
        # Inline equation below Equation 5 Kaminski & Ribe (2001)
        bigi = np.zeros(4)
        for j in range(3):
            for k in range(3):
                bigi[0] += ex[j, k] * acsi[i, 0, j] * acsi[i, 1, k]
                bigi[1] += ex[j, k] * acsi[i, 0, j] * acsi[i, 2, k]
                bigi[2] += ex[j, k] * acsi[i, 2, j] * acsi[i, 1, k]
                bigi[3] += ex[j, k] * acsi[i, 2, j] * acsi[i, 0, k]
        # Equation 5 Kaminski & Ribe (2001)
        indInac, indMin, indInt, indMax = np.argsort(np.abs(bigi / tau))
        gam[indInac], gam[indMax] = 0, 1
        for index in [indMin, indInt]:
            frac = tau[indMax] / tau[index] * bigi[index] / bigi[indMax]
            gam[index] = frac * abs(frac) ** (stressexp - 1)
        # Equation 4 Kaminski & Ribe (2001)
        for j in range(3):
            for k in range(3):
                g[j, k] = 2 * (
                    gam[0] * acsi[i, 0, j] * acsi[i, 1, k]
                    + gam[1] * acsi[i, 0, j] * acsi[i, 2, k]
                    + gam[2] * acsi[i, 2, j] * acsi[i, 1, k]
                    + gam[3] * acsi[i, 2, j] * acsi[i, 0, k]
                )
                g_ens[j, k] = 2 * acsi_ens[i, 2, j] * acsi_ens[i, 0, k]
        # Equation 7 Kaminski & Ribe (2001)
        gamNum = gamNum_ens = gamDen = gamDen_ens = 0
        for j in range(3):
            k = (j + 1) % 3  # Fortran implementation uses j + 2 ???
            gamNum -= (lx[j, k] - lx[k, j]) * (g[j, k] - g[k, j])
            gamNum_ens -= (lx[j, k] - lx[k, j]) * (g_ens[j, k] - g_ens[k, j])
            # Mistake in the equation: kl+1 instead of kk+1 ???
            gamDen -= (g[j, k] - g[k, j]) ** 2
            gamDen_ens -= (g_ens[j, k] - g_ens[k, j]) ** 2
            for L in range(3):
                gamNum += 2 * g[j, L] * lx[j, L]
                gamNum_ens += 2 * g_ens[j, L] * lx[j, L]
                gamDen += 2 * g[j, L] ** 2
                gamDen_ens += 2 * g_ens[j, L] ** 2
        gam0 = gamNum / gamDen
        gam0_ens = gamNum_ens / gamDen_ens / tau_ens**stressexp
        # Equation 22 Kaminski & Ribe (2001)
        for index in [indMin, indInt, indMax]:
            dislDens = tau[index] ** (1.5 - stressexp) * abs(gam[index] * gam0) ** (
                1.5 / stressexp
            )
            rt[i] += dislDens * np.exp(-lamb * dislDens**2)
        rt0_ens = tau_ens ** (1.5 - stressexp) * abs(gam0_ens) ** (1.5 / stressexp)
        rt_ens[i] = rt0_ens * np.exp(-lamb * rt0_ens**2)
        if alpha == 0:
            # Equation 33 Ribe (1992) and Equation 8 Kaminski & Ribe (2001) ???
            rot = rotIni.copy()
            for j in range(3):
                r, s = (j + 1) % 3, (j + 2) % 3
                rot[j] += (lx[s, r] - lx[r, s]) / 2
            rot_ens = rot.copy()
        else:
            # Equation 8 Kaminski & Ribe (2001)
            for j in range(3):
                r, s = (j + 1) % 3, (j + 2) % 3
                rot[j] = (lx[s, r] - lx[r, s] - (g[s, r] - g[r, s]) * gam0) / 2
                rot_ens[j] = (
                    lx[s, r] - lx[r, s] - (g_ens[s, r] - g_ens[r, s]) * gam0_ens
                ) / 2
        # Equation 9 Kaminski & Ribe (2001)
        for p in range(3):
            for q in range(3):
                for r in range(3):
                    for s in range(3):
                        dotacs[i, p, q] += alt[q, r, s] * acsi[i, p, s] * rot[r]
                        dotacs_ens[i, p, q] += (
                            alt[q, r, s] * acsi_ens[i, p, s] * rot_ens[r]
                        )
    # Volume averaged energy
    Emean, Emean_ens = np.sum(odfi * rt), np.sum(odfi_ens * rt_ens)
    # Change of volume fraction by grain boundary migration
    dotodf = Xol * mob * odfi * (Emean - rt)
    dotodf_ens = (1 - Xol) * mob * odfi_ens * (Emean_ens - rt_ens)
    return dotacs, dotacs_ens, dotodf, dotodf_ens


# Calculation of strain along pathlines
def strain(pathTime, pathDense, dictGlobals):
    chi, gridCoords, iDefMech, size = (
        dictGlobals["chi"],
        dictGlobals["gridCoords"],
        dictGlobals["iDefMech"],
        dictGlobals["size"],
    )
    fse = np.identity(3)
    # Uniformly distributed rotations represented as rotation matrices
    acs0 = R.random(size, random_state=1).as_matrix()
    acs, acs_ens = acs0.copy(), acs0.copy()
    odf, odf_ens = np.ones(size) / size, np.ones(size) / size
    for time in reversed(pathTime):
        if isInside(pathDense(time), dictGlobals):
            currTime = time
            break
    while currTime < pathTime[0]:
        currPoint = pathDense(currTime)
        currVel = interpVel(currPoint, dictGlobals)
        L, epsnot = interpVelGrad(currPoint, dictGlobals)
        e = (L + L.transpose()) / 2
        alpha = iDefMech(*currPoint)
        indLeft = [
            np.searchsorted(gridCoords[x], currPoint[x]) for x in range(currPoint.size)
        ]
        gridStep = []
        for coord, ind in zip(gridCoords, indLeft):
            try:
                gridStep.append(coord[ind + 1] - coord[ind])
            except IndexError:
                gridStep.append(coord[ind] - coord[ind - 1])
        dtPathline = min(gridStep) / 4 / norm(currVel, ord=2)
        dtPathline = min(dtPathline, pathTime[0] - currTime)
        # time stepping for LPO calculation
        dt = min(dtPathline, 1e-2 / epsnot)
        # number of iterations in the LPO loop
        nbIter = int(dtPathline / dt)
        # LPO loop on the point on the pathline
        for iter in range(nbIter):
            # CALL 1/4
            dotacs, dotacs_ens, dotodf, dotodf_ens = deriv(
                L / epsnot, e / epsnot, acs, acs_ens, fse, odf, odf_ens, alpha
            )
            kfse1 = np.dot(L, fse) * dt
            kodf1 = dotodf * dt * epsnot
            kodf1_ens = dotodf_ens * dt * epsnot
            kac1 = dotacs * dt * epsnot
            kac1_ens = dotacs_ens * dt * epsnot
            fsei = fse + 0.5 * kfse1
            acsi = np.clip(acs + 0.5 * kac1, -1, 1)
            acsi_ens = np.clip(acs_ens + 0.5 * kac1_ens, -1, 1)
            odfi = np.clip(odf + 0.5 * kodf1, 0, None)
            odfi /= odfi.sum()
            odfi_ens = np.clip(odf_ens + 0.5 * kodf1_ens, 0, None)
            odfi_ens /= odfi_ens.sum()
            # CALL 2/4
            dotacs, dotacs_ens, dotodf, dotodf_ens = deriv(
                L / epsnot, e / epsnot, acsi, acsi_ens, fse, odfi, odfi_ens, alpha
            )
            kfse2 = np.dot(L, fsei) * dt
            kodf2 = dotodf * dt * epsnot
            kodf2_ens = dotodf_ens * dt * epsnot
            kac2 = dotacs * dt * epsnot
            kac2_ens = dotacs_ens * dt * epsnot
            fsei = fse + 0.5 * kfse2
            acsi = np.clip(acs + 0.5 * kac2, -1, 1)
            acsi_ens = np.clip(acs_ens + 0.5 * kac2_ens, -1, 1)
            odfi = np.clip(odf + 0.5 * kodf2, 0, None)
            odfi /= odfi.sum()
            odfi_ens = np.clip(odf_ens + 0.5 * kodf2_ens, 0, None)
            odfi_ens /= odfi_ens.sum()
            # CALL 3/4
            dotacs, dotacs_ens, dotodf, dotodf_ens = deriv(
                L / epsnot, e / epsnot, acsi, acsi_ens, fse, odfi, odfi_ens, alpha
            )
            kfse3 = np.dot(L, fsei) * dt
            kodf3 = dotodf * dt * epsnot
            kodf3_ens = dotodf_ens * dt * epsnot
            kac3 = dotacs * dt * epsnot
            kac3_ens = dotacs_ens * dt * epsnot
            fsei = fse + kfse3
            acsi = np.clip(acs + kac3, -1, 1)
            acsi_ens = np.clip(acs_ens + kac3_ens, -1, 1)
            odfi = np.clip(odf + kodf3, 0, None)
            odfi /= odfi.sum()
            odfi_ens = np.clip(odf_ens + kodf3_ens, 0, None)
            odfi_ens /= odfi_ens.sum()
            # CALL 4/4
            dotacs, dotacs_ens, dotodf, dotodf_ens = deriv(
                L / epsnot, e / epsnot, acsi, acsi_ens, fse, odfi, odfi_ens, alpha
            )
            kfse4 = np.dot(L, fsei) * dt
            kodf4 = dotodf * dt * epsnot
            kodf4_ens = dotodf_ens * dt * epsnot
            kac4 = dotacs * dt * epsnot
            kac4_ens = dotacs_ens * dt * epsnot
            fse += (kfse1 / 2 + kfse2 + kfse3 + kfse4 / 2) / 3
            acs = np.clip(acs + (kac1 / 2 + kac2 + kac3 + kac4 / 2) / 3, -1, 1)
            acs_ens = np.clip(
                acs_ens + (kac1_ens / 2 + kac2_ens + kac3_ens + kac4_ens / 2) / 3, -1, 1
            )
            odf += (kodf1 / 2 + kodf2 + kodf3 + kodf4 / 2) / 3
            odf_ens += (kodf1_ens / 2 + kodf2_ens + kodf3_ens + kodf4_ens / 2) / 3
            odf /= odf.sum()
            odf_ens /= odf_ens.sum()
            # Grain-boundary sliding
            mask = odf < chi / size
            mask_ens = odf_ens < chi / size
            acs[mask, :, :] = acs0[mask, :, :]
            acs_ens[mask_ens, :, :] = acs0[mask_ens, :, :]
            odf[mask] = chi / size
            odf_ens[mask_ens] = chi / size
            odf /= odf.sum()
            odf_ens /= odf_ens.sum()
        currTime += nbIter * dt
    return fse, acs, acs_ens, odf, odf_ens


# Calculates elastic tensor Cav_{ijkl} for olivine
@jit(nopython=True)
def voigt(acs, acs_ens, odf, odf_ens):
    C0, C0_ens = np.empty((3, 3, 3, 3)), np.empty((3, 3, 3, 3))
    Cav, Sav = np.zeros((3, 3, 3, 3)), np.empty((6, 6))
    # Indices to form Cijkl from Sij
    ijkl = np.array([[0, 5, 4], [5, 1, 3], [4, 3, 2]], dtype=np.int16)
    # Elastic tensors
    for p in range(3):
        for q in range(3):
            i = ijkl[p, q]
            for r in range(3):
                for s in range(3):
                    j = ijkl[r, s]
                    C0[p, q, r, s], C0_ens[p, q, r, s] = S0[i, j], S0_ens[i, j]
    for nu in range(size):
        C0rot = rotateTens(C0, acs[nu, ...].transpose())
        C0rot_ens = rotateTens(C0_ens, acs_ens[nu, ...].transpose())
        Cav += C0rot * odf[nu] * Xol + C0rot_ens * odf_ens[nu] * (1 - Xol)
    # Average stiffness matrix
    for i in range(6):
        p, q = (i + i // 3) % 3, (i + 2 * (i // 3)) % 3
        for j in range(6):
            r, s = (j + j // 3) % 3, (j + 2 * (j // 3)) % 3
            Sav[i, j] = Cav[p, q, r, s]
    return Sav


# Calculates GOL parameter at grid point
def pipar(currPoint, dictGlobals):
    # Calculates ISA Orientation at grid point
    def isacalc(L, veloc):
        nonlocal GOL
        # Kaminski, Ribe & Browaeys (2004): Appendix B
        evals = eigvalsh(L)
        if np.sum(abs(evals) < 1e-9) >= 2 - veloc.size % 2:
            assert (
                abs(evals[0] * evals[1] + evals[0] * evals[2] + evals[1] * evals[2])
                < 1e-9
            )
            return veloc / norm(veloc, ord=2)
        else:
            ind = np.argsort(abs(evals))[-1]
            if np.isreal(evals[ind]):
                a, b = (ind + 1) % 3, (ind + 2) % 3
                Id = np.identity(3)
                Fa = (
                    np.dot(L - evals[a] * Id, L - evals[b] * Id)
                    / (evals[ind] - evals[a])
                    / (evals[ind] - evals[b])
                )
                Ua = np.dot(np.transpose(Fa), Fa)
                return eigh(Ua)[1][:: 4 - veloc.size, -1]
            else:
                GOL = -1
                return np.zeros(veloc.size)

    GOL = 10
    # local velocity vector and orientation of infinite strain axis
    veloc = interpVel(currPoint, dictGlobals)
    dt = 1e-2 / norm(veloc, ord=2)
    # previous point on the streamline
    prevPoint = currPoint - dt * veloc
    if isInside(prevPoint, dictGlobals) is False:
        return -10
    veloc = interpVel(prevPoint, dictGlobals)
    # epsnot: reference strain rate
    L, epsnot = interpVelGrad(prevPoint, dictGlobals)
    L /= epsnot
    # calculation of the ISA
    isa = isacalc(L, veloc)
    # angle between ISA and flow direction
    thetaISA = np.arccos(np.sum(veloc / norm(veloc, ord=2) * isa))
    # next point on the streamline
    nextPoint = currPoint + dt * veloc
    if isInside(nextPoint, dictGlobals) is False:
        return -10
    veloc = interpVel(nextPoint, dictGlobals)
    L, epsnot = interpVelGrad(nextPoint, dictGlobals)
    L /= epsnot
    # calculation of the ISA
    isa = isacalc(L, veloc)
    L, epsnot = interpVelGrad(currPoint, dictGlobals)
    # angle between ISA and flow direction
    thetaISA = abs(thetaISA - np.arccos(np.sum(veloc / norm(veloc, ord=2) * isa)))
    if thetaISA > np.pi:
        thetaISA -= np.pi
    return min(GOL, thetaISA / 2 / dt / epsnot)


# Determines a pathline given a position and a velocity field
def pathline(currPoint, dictGlobals):
    def ivpFunc(time, pointCoords, dictGlobals):
        if isInside(pointCoords, dictGlobals):
            return interpVel(pointCoords, dictGlobals)
        else:
            return np.zeros(pointCoords.shape)

    def ivpJac(time, pointCoords, dictGlobals):
        if isInside(pointCoords, dictGlobals):
            L = interpVelGrad(pointCoords, dictGlobals, True)
            if pointCoords.size == 2:
                return np.array([[L[0, 0], L[0, 2]], [L[2, 0], L[2, 2]]])
            else:
                return L
        else:
            return np.zeros((pointCoords.size, pointCoords.size))

    def maxStrain(time, pointCoords, dictGlobals):
        nonlocal eventTime, eventTimePrev, eventStrain, eventStrainPrev, eventFlag
        if eventFlag:
            return (eventStrain if time == eventTime else eventStrainPrev) - 10
        elif isInside(pointCoords, dictGlobals):
            L, epsnot = interpVelGrad(pointCoords, dictGlobals)
            eventStrainPrev = eventStrain
            eventStrain += abs(time - eventTime) * epsnot
            if eventStrain >= 10:
                eventFlag = True
            eventTimePrev = eventTime
            eventTime = time
            return eventStrain - 10
        else:
            return 0

    maxStrain.terminal = True
    eventStrain = eventTime = 0
    eventStrainPrev = eventTimePrev = None
    eventFlag = False
    with catch_warnings():
        simplefilter("ignore", category=UserWarning)
        sol = solve_ivp(
            ivpFunc,
            [0, -100e6 * 365.25 * 8.64e4],
            currPoint,
            method="RK45",
            first_step=1e10,
            max_step=np.inf,
            t_eval=None,
            events=[maxStrain],
            args=(dictGlobals,),
            dense_output=True,
            jac=ivpJac,
            atol=1e-8,
            rtol=1e-5,
        )
    return sol.t, sol.sol


# Checks if a point lies within the numerical domain
def isInside(point, dictGlobals):
    gridMax, gridMin = dictGlobals["gridMax"], dictGlobals["gridMin"]
    inside = True
    for coord, minBound, maxBound in zip(point, gridMin, gridMax):
        if coord < minBound or coord > maxBound:
            inside = False
            break
    return inside


def DRex(locInd, dictGlobals):
    gridCoords = dictGlobals["gridCoords"]
    # Initial location of the grid point
    currCoords = np.array(
        [(coord[ind] + coord[ind + 1]) / 2 for coord, ind in zip(gridCoords, locInd)]
    )
    # Grain Orientation Lag
    GOL = pipar(currCoords, dictGlobals)
    # Backward calculation of the pathline for each tracer
    pathTime, pathDense = pathline(currCoords, dictGlobals)
    # Inward calculation of the LPO | Random initial LPO
    Fij, acs, acs_ens, odf, odf_ens = strain(pathTime, pathDense, dictGlobals)
    # Left-stretch tensor for FSE calculation
    LSij = np.dot(Fij, np.transpose(Fij))
    eigval, eigvect = eigh(LSij)
    # pick up the orientation of the long axis of the FSE
    if len(locInd) == 2:
        phi_fse = np.arctan2(eigvect[-1, -1], eigvect[0, -1])
    else:
        phi_fse = np.arctan2(eigvect[-1, -1], norm(eigvect[:-1, -1])) * np.sign(
            eigvect[:-1, -1].prod()
        )
    # natural strain = ln(a / c) where a is the long axis = max(eigval) ** 0.5
    ln_fse = np.log(eigval[-1] / eigval[0]) / 2
    # Cijkl tensor (using Voigt average)
    Sav = voigt(acs, acs_ens, odf, odf_ens)
    # percentage anisotropy and orientation of hexagonal symmetry axis
    perc_a, phi_a, radani, percani, azimuth = decsym(Sav)
    return (locInd[::-1], radani, percani, GOL, ln_fse, phi_fse, phi_a, perc_a, azimuth)


def main(inputArgs):
    if isinstance(inputArgs, list):
        inputArgs = args

    inputExt = inputArgs.input.split(".")[-1]
    if inputExt == "vtu":
        vtkReader = vtkXMLUnstructuredGridReader()
    elif inputExt == "pvtu":
        vtkReader = vtkXMLPUnstructuredGridReader()
    else:
        raise RuntimeError(
            """Please either provide a supported file format or implement the
    reader your require."""
        ) from None

    vtkReader.SetFileName(inputArgs.input)
    vtkReader.Update()
    vtkOut = vtkReader.GetOutput()
    vtkData = vtkOut.GetPoints().GetData()
    coords = np.array(
        [vtkData.GetTuple3(tup) for tup in range(vtkData.GetNumberOfTuples())]
    )
    if np.amin(coords[:, 1]) == np.amax(coords[:, 1]):
        raise RuntimeError("""Please either provide a 2-D or 3-D dataset.""") from None
    elif np.amin(coords[:, 2]) == np.amax(coords[:, 2]):
        dim = 2
        coords[:, 1] = np.amax(coords[:, 1]) - coords[:, 1]
    else:
        dim = 3
        coords[:, 2] = np.amax(coords[:, 2]) - coords[:, 2]
    vtkScalars = vtkOut.GetPointData().GetScalars("DeformationMechanism")
    defMech = np.array(
        [vtkScalars.GetTuple1(tup) for tup in range(vtkScalars.GetNumberOfTuples())]
    )
    vtkScalars = vtkOut.GetPointData().GetScalars("Velocity")
    vel = np.array(
        [vtkScalars.GetTuple3(tup) for tup in range(vtkScalars.GetNumberOfTuples())]
    )
    vtkArray = vtkOut.GetPointData().GetArray("VelocityGradient")
    velGrad = np.array(
        [vtkArray.GetTuple9(tup) for tup in range(vtkArray.GetNumberOfTuples())]
    )
    del vtkReader, vtkData, vtkScalars, vtkArray

    dictGlobals = {
        "chi": chi,
        "gridMin": gridMin,
        "gridMax": gridMax,
        "size": size,
        "gridNodes": gridNodes,
        "gridCoords": gridCoords,
    }

    if dim == 2:
        interpObj = [["iVelX", "iVelZ"], ["iLxx", "iLzx", "iLxz", "iLzz"]]
        if inputArgs.mpl:
            triIDs = []
            for i in range(vtkOut.GetNumberOfCells()):
                assert vtkOut.GetCell(i).GetNumberOfPoints() == 3
                IDsList = vtkOut.GetCell(i).GetPointIds()
                triIDs.append(
                    [IDsList.GetId(j) for j in range(IDsList.GetNumberOfIds())]
                )
            tri = Triangulation(coords[:, 0], coords[:, 1], triangles=triIDs)
            for objList, fieldVar in zip(interpObj, [vel, velGrad]):
                for i, objName in enumerate(objList):
                    dictGlobals[objName] = CubicTriInterpolator(
                        tri, fieldVar[:, i % 2 + i // 2 * 3], kind=inputArgs.mpl
                    )
        else:
            tri = Delaunay(coords[:, :-1])
            for objList, fieldVar in zip(interpObj, [vel, velGrad]):
                for i, objName in enumerate(objList):
                    dictGlobals[objName] = CloughTocher2DInterpolator(
                        tri, fieldVar[:, i % 2 + i // 2 * 3]
                    )
        dictGlobals["iDefMech"] = NearestNDInterpolator(coords[:, :-1], defMech)
        del tri, vtkOut
    else:
        interpObj = [
            ["iVelX", "iVelY", "iVelZ"],
            ["iLxx", "iLyx", "iLzx", "iLxy", "iLyy", "iLzy", "iLxz", "iLyz", "iLzz"],
        ]
        for objList, fieldVar in zip(interpObj, [vel, velGrad]):
            for i, objName in enumerate(objList):
                dictGlobals[objName] = NearestNDInterpolator(coords, fieldVar[:, i])
        dictGlobals["iDefMech"] = NearestNDInterpolator(coords, defMech)

    outVar = [
        "radani",
        "percani",
        "GOL",
        "ln_fse",
        "phi_fse",
        "phi_a",
        "perc_a",
        "azimuth",
    ]
    varDict = {}
    if inputArgs.restart:
        varDict.update(np.load(inputArgs.restart))
    else:
        arrDim = [x - 1 for x in reversed(gridNodes)]
        for var in outVar:
            varDict[var] = np.zeros(arrDim)
        varDict["indArr"] = np.zeros(arrDim[::-1])
        varDict["nodesComplete"] = 0

    if inputArgs.charm:  # Charm4Py
        nBatch = np.ceil(np.sum(varDict["indArr"] == 0) / 6e4).astype(int)
        for batch in range(nBatch):
            nodes2do = np.asarray(varDict["indArr"] == 0).nonzero()
            futures = charm.pool.map_async(
                partial(DRex, dictGlobals=dictGlobals),
                list(zip(*[nodes[:60_000] for nodes in nodes2do])),
                multi_future=True,
            )
            for future in charm.iwait(futures):
                output = future.get()
                for i, var in enumerate(outVar):
                    varDict[var][output[0]] = output[i + 1]
                varDict["indArr"][output[0][::-1]] = 1
                varDict["nodesComplete"] += 1
                if varDict["nodesComplete"] % checkpoint == 0:
                    np.savez(
                        f"PyDRex{dim}D_{name}_NumpyCheckpoint_"
                        f"{varDict['nodesComplete']}",
                        **varDict,
                    )
    elif inputArgs.ray:  # Ray
        if inputArgs.redis_pass:
            # Cluster execution
            ray.init(address="auto", _redis_password=inputArgs.redis_pass)
        else:
            # Single machine | Set local_mode to True to force single process
            ray.init(num_cpus=inputArgs.cpus, local_mode=False)

        dictGlobalsID = ray.put(dictGlobals)

        nBatch = np.ceil(np.sum(varDict["indArr"] == 0) / 6e4).astype(int)
        for batch in range(nBatch):
            nodes2do = np.asarray(varDict["indArr"] == 0).nonzero()
            futures = [
                DRex.remote(i, dictGlobalsID)
                for i in zip(*[nodes[:60_000] for nodes in nodes2do])
            ]
            while len(futures):
                readyIds, futures = ray.wait(
                    futures, num_returns=min([checkpoint, len(futures)])
                )
                for output in ray.get(readyIds):
                    for i, var in enumerate(outVar):
                        varDict[var][output[0]] = output[i + 1]
                    varDict["indArr"][output[0][::-1]] = 1
                    varDict["nodesComplete"] += 1
                del readyIds
                np.savez(
                    f"PyDRex{dim}D_{name}_NumpyCheckpoint_"
                    f"{varDict['nodesComplete']}",
                    **varDict,
                )

        ray.shutdown()
    else:  # Multiprocessing
        if __name__ == "__main__":
            with Pool(processes=inputArgs.cpus) as pool:
                for output in pool.imap_unordered(
                    partial(DRex, dictGlobals=dictGlobals),
                    zip(*np.asarray(varDict["indArr"] == 0).nonzero()),
                ):
                    for i, var in enumerate(outVar):
                        varDict[var][output[0]] = output[i + 1]
                    varDict["indArr"][output[0][::-1]] = 1
                    varDict["nodesComplete"] += 1
                    if varDict["nodesComplete"] % checkpoint == 0:
                        np.savez(
                            f"PyDRex{dim}D_{name}_NumpyCheckpoint_"
                            f"{varDict['nodesComplete']}",
                            **varDict,
                        )

    np.savez(f"PyDRex{dim}D_{name}_Complete", **varDict)

    end = perf_counter()
    hours = int((end - begin) / 3600)
    minutes = int((end - begin - hours * 3600) / 60)
    seconds = end - begin - hours * 3600 - minutes * 60
    print(f"{hours}:{minutes}:{seconds}")

    if inputArgs.charm:
        exit()


begin = perf_counter()

parser = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description="""
Python 3 implementation of the seismic anisotropy software D-Rex. Parallel
execution is achieved through multiprocessing; for multi-node parallelism
enable either Ray or Charm4Py. The module DRexParam.py must be in the same
directory.""",
)
parser.add_argument("input", help="input file (expects a vtkUnstructuredGrid)")
parser.add_argument(
    "--charm",
    action="store_true",
    help="""enables parallel
                    execution through Charm4Py""",
)
parser.add_argument(
    "--cpus",
    help="""Number of CPUs requested; required when
                    executing on a single-node environment (except for
                    Charm4Py)""",
    metavar="",
    type=int,
)
parser.add_argument(
    "--mpl",
    help="""Must be either 'geom' or 'min_E'; only
                    applies to 2-D simulations for which a mesh triangulation
                    already exists. If provided, velocity and velocity gradient
                    fields are interpolated using Matplotlib
                    CubicTriInterpolator instead of Scipy
                    CloughTocher2DInterpolator.""",
    metavar="",
)
parser.add_argument(
    "--ray",
    action="store_true",
    help="""enables parallel
                    execution through Ray""",
)
parser.add_argument(
    "--redis-pass",
    help="""Redis password; required when using
                    Ray on a multi-node environment""",
    metavar="",
)
parser.add_argument(
    "--restart",
    help="""checkpoint file to restart from
                    (expects a Numpy NpzFile)""",
    metavar="",
)
args = parser.parse_args()

if args.charm and args.cpus:
    print(
        """Warning: --cpus is redundant when using Charm4Py; consider
          removing it."""
    )

if (args.ray and args.redis_pass) and args.cpus:
    print(
        """Warning: --cpus is redundant when using Ray in a multi-node
          envrionment; consider removing it."""
    )

if args.charm or args.ray:
    assert (args.charm and args.ray) is False

if args.charm:
    from charm4py import charm
elif args.ray:
    import ray

    DRex = ray.remote(DRex)
else:
    from multiprocessing import Pool

if args.ray is False:
    from functools import partial

if args.mpl:
    from matplotlib.tri import CubicTriInterpolator, Triangulation
else:
    from scipy.interpolate import CloughTocher2DInterpolator
    from scipy.spatial import Delaunay

if args.charm:
    charm.start(main)
else:
    main(args)
